import abc
import typing

from goats.core import algebraic
from goats.core import metadata


# This should support algebraic operations on metadata attributes without
# knowledge of the type of interface. Some guiding use cases:
# - If the interface is an instance of `variable.Quantity`, `implement` should
#   evaluate it with `func` (see `measurable.Quantified.implement`).
# - If the interface is an instance of `computable.Quantity`, `implement` should
#   do nothing.
class Base(algebraic.Quantity):
    """The base quantity."""
    def __init__(self, __interface) -> None:
        self._interface = __interface
        self._meta = None

    @property
    def meta(self):
        """This quantity's metadata attributes."""
        if self._meta is None:
            self._meta = metadata.OperatorFactory(type(self))
        return self._meta

    def __eq__(self, other) -> bool:
        """Called for self == other."""
        for name in self.meta.parameters:
            v = getattr(self, name)
            if hasattr(other, name) and getattr(other, name) != v:
                return False
        return True

    def implement(self, func: typing.Callable, mode: str, *others, **kwargs):
        """Implement a standard operation."""
        name = func.__name__
        if mode == 'cast':
            return self.compute(func, mode)
        if mode == 'arithmetic':
            data = self.compute(func, mode, **kwargs)
            meta = self.meta[name].evaluate(self, **kwargs)
            return type(self)(data, **meta)
        operands = [self] + list(others)
        args = [
            self.compute(func, mode, **kwargs)
            if isinstance(i, Base) else i
            for i in operands
        ]
        if mode == 'comparison':
            self.meta.check(self, *others)
            return self.compute(func, mode, *args)
        if mode == 'forward':
            data = self.compute(func, mode, *args, **kwargs)
            meta = self.meta[name].evaluate(*operands, **kwargs)
            return type(self)(data, **meta)
        if mode == 'reverse':
            data = self.compute(func, mode, *reversed(args), **kwargs)
            meta = self.meta[name].evaluate(*reversed(operands), **kwargs)
            return type(self)(data, **meta)
        if mode == 'inplace':
            data = self.compute(func, mode, *args, **kwargs)
            meta = self.meta[name].evaluate(*operands, **kwargs)
            self._data = data
            for k, v in meta:
                setattr(self, k, v)
            return self
        raise ValueError(f"Unknown operator mode {mode!r}")

    @abc.abstractmethod
    def compute(self, func: typing.Callable, mode: str, *args, **kwargs):
        """Evaluate this quantity's data interface."""
        pass





import collections.abc
import configparser
import json
import os
import pathlib

from goats.core import iotools


# read version from installed package
from importlib.metadata import version
__version__ = version("goats")


class Environment(collections.abc.Mapping):
    """A collection of environmental settings."""

    def __init__(self, name: str) -> None:
        self.name = name
        """The name of the observing package to select."""
        self._package = f"{__package__}.{self.name}"
        home = pathlib.Path('~').expanduser()
        paths = [
            pathlib.Path.cwd(), # The current working directory
            home, # The user's home directory
            home / '.config', # Linux standard (local)
            '/etc/goats', # Linux standard (global)
            os.environ.get('GOATS_INI'), # A known environment variable
            pathlib.Path(__file__).parent, # The package top
        ]
        config = configparser.ConfigParser()
        path = iotools.search(paths, 'goats.ini')
        config.read(iotools.ReadOnlyPath(path))
        self._config = config[self.name]
        self.path = path

    def __len__(self) -> int:
        """The number of available parameter values."""
        return len(self._config)

    def __iter__(self):
        """Iterate over available parameter values."""
        yield from self._config

    def __getitem__(self, key: str):
        """Access parameter values by mapping key."""
        if key in self._config:
            return self._config[key]
        raise KeyError(
            f"{self._package} has no value for {key!r}"
        ) from None

    def __str__(self) -> str:
        return json.dumps(
            dict(self._config),
            indent=4,
            sort_keys=True,
        )

    def __repr__(self) -> str:
        return f"{self._package}({self.path}):\n{self}"



import typing

import numpy
import numpy.typing
from scipy.interpolate import interp1d

from goats.core import numerical
from goats.core import variable


def restrict_coordinate(
    array: numpy.ndarray,
    target: float,
    axis: typing.SupportsIndex,
) -> range:
    """Restrict `array` to `target` along `axis`."""
    bounds = numerical.get_bounding_indices(array, target, axis=axis)
    if bounds.ndim == 1:
        return range(bounds[0], bounds[1]+1)
    if bounds.ndim == 2:
        lower = min(bounds[:, 0])
        upper = max(bounds[:, 1])
        return range(lower, upper+1)
    # NOTE: We can get `lower` and `upper` for N-D arrays by replacing `:` with
    # `...`. The problem is that they will be (N-1)-D arrays that we will then
    # need to split into individual index vectors before returning.
    raise NotImplementedError(f"Can't operate on {bounds.ndim}-D array")


class Restriction:
    """A subset of indices with which to create subsets of arrays."""

    def __init__(
        self,
        rule: typing.Callable[..., typing.Iterable[typing.SupportsIndex]],
        *args,
        **kwargs,
    ) -> None:
        """
        Create a new restriction.

        Parameters
        ----------
        rule : callable
            The rule by which to create a subset of array indices. This must be
            a callable object that returns an iterable of objects to be used as
            indices (e.g., `range`).

        *args
            Positional arguments to pass to `rule`.

        **kwargs
            Keyword arguments to pass to `rule`.
        """
        self.indices = rule(*args, **kwargs)

    def apply(self, array: numpy.typing.ArrayLike, axis: int=None):
        """Restrict `array` to a subset along `axis`.
        
        Parameters
        ----------
        array : array-like
            The array from which to extract a subarray corresponding to the
            restricted indices.

        axis : int, default=-1
            The axis of `array` to restrict.

        Returns
        -------
        array-like
            The subarray.
        """
        if axis is None:
            axis = -1
        permuted = numpy.moveaxis(array, axis, 0)
        result = permuted[self.indices, ...]
        return numpy.moveaxis(result, 0, axis)


class Quantity(variable.Quantity):
    """A variable quantity, interpolated over one or more axes."""

    def interpolate(
        self,
        coordinates: typing.Iterable[typing.Dict[str, typing.Any]],
    ) -> variable.Quantity:
        """Internal interpolation logic."""
        for coordinate in coordinates:
            self._array = self._interpolate_coordinate(
                coordinate['targets'],
                coordinate['reference'],
                axis=coordinate.get('axis'),
            )
        return self

    def _interpolate_coordinate(
        self,
        targets: numpy.ndarray,
        reference: variable.Quantity,
        axis: int=None,
    ) -> numpy.ndarray:
        """Interpolate a variable array based on a known coordinate."""
        array = numpy.array(self)
        indices = (self.axes.index(d) for d in reference.axes)
        dst, src = zip(*enumerate(indices))
        reordered = numpy.moveaxis(array, src, dst)
        interpolated = _apply_interpolation(
            reordered,
            numpy.array(reference),
            targets,
            axis=axis,
        )
        return numpy.moveaxis(interpolated, dst, src)


def _apply_interpolation(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    targets: typing.Iterable[float],
    axis: int=None,
) -> numpy.ndarray:
    """Interpolate `array` to target values over `coordinate`."""
    interpolated = [
        _apply_interp1d(array, reference, target, axis=axis)
        for target in targets
    ]
    if reference.ndim == 2:
        return numpy.swapaxes(interpolated, 0, 1)
    return numpy.array(interpolated)


def _apply_interp1d(
    array: numpy.ndarray,
    reference: numpy.ndarray,
    target: float,
    axis: int=None,
) -> typing.List[float]:
    """Interpolate data to `target` along the leading axis."""
    if target in reference:
        idx = numerical.find_nearest(reference, target).index
        return array[idx]
    if axis is not None:
        restriction = Restriction(
            restrict_coordinate,
            reference,
            target,
            axis=axis,
        )
        ref = restriction.apply(reference, axis=axis)
        arr = restriction.apply(array, axis=axis)
    else:
        ref = reference
        arr = array
    if reference.ndim == 2:
        interps = [interp1d(x, y, axis=0) for x, y in zip(ref, arr)]
        return [interp(target) for interp in interps]
    interp = interp1d(ref, arr, axis=0)
    return interp(target)





import collections.abc
import contextlib
import numbers
import typing

import numpy
import numpy.typing

from goats.core import algebraic
from goats.core import fundamental
from goats.core import iterables
from goats.core import measurable
from goats.core import metadata
from goats.core import metric


Instance = typing.TypeVar('Instance', bound='Quantity')


class Quantity(measurable.Quantity, metadata.NameMixin):
    """A measurable quantity with a name."""

    @typing.overload
    def __init__(
        self: Instance,
        __data: algebraic.Real,
        *,
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> None: ...

    def __init__(self, __data, **meta) -> None:
        """Initialize this instance from arguments or an existing instance."""
        super().__init__(__data, **meta)
        parsed = self.parse_attrs(__data, meta, name='')
        self._name = metadata.Name(parsed['name'])
        self.meta.register('name')
        if self._name:
            self.display.register('name')
            self.display['__str__'].insert(0, "'{name}':")
            self.display['__repr__'].insert(2, "name='{name}'")


class Scalar(Quantity, algebraic.Scalar):
    """A single-valued named quantity."""

    @typing.overload
    def __init__(
        self: Instance,
        __data: numbers.Real,
        *,
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> None: ...

    def __init__(self, __data, **meta) -> None:
        super().__init__(float(__data), **meta)

    def __iter__(self):
        """Explicitly suppress iteration.
        
        This is necessary because `~measurable.Quantity` uses `__getitem__` for
        unit updates. At the same time, built-in code will attempt to use the
        old-style iterator protocol if it doesn't find an `__iter__` method,
        which includes calling `__getitem__` (cf.
        https://docs.python.org/3/library/collections.abc.html). Defining
        `__iter__` as not implemented circumvents that problem.
        """
        raise TypeError(f"Can't iterate over {type(self)!r}") from None


class Constants(collections.abc.Mapping):
    """Fundamental physical constants in a given metric system."""
    def __init__(self, system: str) -> None:
        self.system = system.lower()
        self._mapping = fundamental.CONSTANTS.copy()

    def __len__(self) -> int:
        """The number of defined constants."""
        return len(self._mapping)

    def __iter__(self) -> typing.Iterator:
        """Iterate over defined constants."""
        return iter(self._mapping)

    def __getitem__(self, name: str):
        """Create the named constant or raise an error."""
        if name in self._mapping:
            found = self._get_attributes(name)
            return Scalar(found['value'], unit=found['unit'])
        raise KeyError(name)

    def _get_attributes(self, name: str) -> dict:
        """Get the value and unit for a named constant, if possible."""
        definition = self._mapping[name]
        if 'all' in definition:
            return {'value': definition['all'], 'unit': None}
        if this := definition.get(self.system):
            return this
        raise ValueError(f"Unknown constant: {name!r}")

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"{self.__class__.__qualname__}({self.system})"


class Vector(Quantity):
    """A multi-valued named quantity.

    Notes
    -----
    This class converts `data` into a `numpy.ndarray` via a `list` during
    instantiation. It may therefore become a bottleneck for large 1-D objects
    and may produce unexpected results for higher-dimension objects. In those
    cases, an array-like class derived from `~measurable.Quantified` that
    incorporates native `numpy` operators may be more appropriate.
    """

    @typing.overload
    def __init__(
        self: Instance,
        __data: typing.Union[algebraic.Real, numpy.typing.ArrayLike],
        *,
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: Instance,
        __data: measurable.Measurement,
        *,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: Instance,
        scalar: Scalar,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> None: ...

    def __init__(self, __data, **meta) -> None:
        if isinstance(__data, measurable.Measurement):
            meta = {'unit': __data.unit}
            __data = __data.values
        elif isinstance(__data, Scalar):
            meta = {k: getattr(__data, k) for k in ('unit', 'name')}
            __data = __data.data
        array = numpy.asfarray(list(iterables.whole(__data)))
        super().__init__(array, **meta)

    def __iter__(self) -> typing.Iterator:
        """Called for iter(self)."""
        return iter(self.data)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.data)

    def __getitem__(self, index):
        """Called for index-based value access."""
        if isinstance(index, str):
            return super().__getitem__(index)
        if isinstance(index, typing.SupportsIndex) and index < 0:
            index += len(self)
        values = self.data[index]
        iter_values = isinstance(values, typing.Iterable)
        unit = self.unit
        return (
            [Scalar(value, unit=unit) for value in values] if iter_values
            else Scalar(values, unit=unit)
        )

    def __eq__(self, other: typing.Any):
        """True if two instances have the same data and attributes."""
        if not isinstance(other, type(self)):
            return NotImplemented
        if not self.unit == other.unit:
            return False
        return numpy.array_equal(other, self)


def scalar(this) -> Scalar:
    """Make sure `this` is a `~physical.Scalar`."""
    if isinstance(this, Scalar):
        return this
    if isinstance(this, Vector) and len(this) == 1:
        return this[0]
    if isinstance(this, measurable.Measurement) and len(this) == 1:
        return Scalar(this.values[0], unit=this.unit)
    measured = measurable.measure(this)
    if len(measured) > 1:
        raise ValueError(
            "Can't create a scalar from a multi-valued quantity"
        ) from None
    return scalar(measured)


IndexLike = typing.TypeVar(
    'IndexLike',
    typing.Iterable[int],
    slice,
    type(Ellipsis),
)
IndexLike = typing.Union[typing.Iterable[int], slice, type(Ellipsis)]


Instance = typing.TypeVar('Instance', bound='Array')


class Array(numpy.lib.mixins.NDArrayOperatorsMixin, Quantity):
    """Base class for array-like objects."""

    def __new__(cls, *args, **kwargs):
        """Create an instance of the appropriate object."""
        if len(args) == 1:
            data = args[0]
            if isinstance(data, numbers.Number):
                return Scalar(data, **kwargs)
        return super().__new__(cls)

    def __init__(self, *args, **kwargs) -> None:
        """Initialize an array from arguments."""
        super().__init__(*args, **kwargs)
        self._scale = 1.0
        self._full_array = None
        self._ndim = None
        self._shape = None
        self.display.register(data='_get_data_type')

    @property
    def data(self):
        """The initial data.
        
        This attribute represents the data-containing object used to create this
        instance. It does not necessarily represent the current array data,
        which may differ as a result of changing the instance unit, or due to
        other allowed updates defined on subclasses.
        """
        return super().data

    def _get_data_type(self):
        """Internal helper for displaying the type of data object."""
        return self.data.__class__

    def __measure__(self):
        """Create a measurement from this array's data and unit."""
        # NOTE: This may produce unexpected results when `self.ndim` > 1.
        return measurable.Measurement(self._array, self.unit)

    def __eq__(self, other: typing.Any):
        """True if two instances have the same data and attributes."""
        if not isinstance(other, type(self)):
            return NotImplemented
        if not self.unit == other.unit:
            return False
        return numpy.array_equal(other, self)

    def __len__(self):
        """Called for len(self)."""
        try:
            method = self._get_base_attr('__len__')
            return method()
        except AttributeError:
            return len(self.data)

    @property
    def ndim(self) -> int:
        """The number of dimensions in this array."""
        if self._ndim is None:
            self._ndim = self._get_base_attr('ndim')
        return self._ndim

    @property
    def shape(self) -> int:
        """The length of each dimension in this array."""
        if self._shape is None:
            self._shape = self._get_base_attr('shape')
        return self._shape

    def _get_base_attr(self, name: str):
        """Helper method to efficiently access underlying attributes.

        This method will first search the underlying data object for the named
        attribute, to take advantage of viewers that provide metadata without
        loading the full dataset. If that search fails, this method will attempt
        to retrieve the named attribute from the actual dataset array, which may
        require loading from disk.
        """
        for this in ['data', '_array']:
            attr = getattr(self, this)
            with contextlib.suppress(AttributeError):
                if hasattr(attr, name):
                    # avoid false negative when value == 0
                    return getattr(attr, name)
        raise AttributeError(f"Could not find an attribute named {name!r}")

    _builtin = (int, slice, type(...), type(None))

    @typing.overload
    def __getitem__(
        self: Instance,
        unit: typing.Union[str, metric.Unit],
    ) -> Instance: ...

    @typing.overload
    def __getitem__(
        self,
        *args: IndexLike,
    ) -> typing.Union[Scalar, 'Array', Instance]: ...

    def __getitem__(self, *args):
        """Create a new instance from a subset of data."""
        if len(args) == 1 and isinstance(args[0], metadata.UnitLike):
            return super().__getitem__(args[0])
        unwrapped = iterables.unwrap(args)
        if iterables.hastype(unwrapped, self._builtin, tuple, strict=True):
            return self._subscript_standard(unwrapped)
        return self._subscript_custom(unwrapped)

    def apply_unit(self, unit: metadata.UnitLike):
        # Create a copy of this instance.
        new = self._copy_with(unit=unit)
        # Update the new instance's internal `scale` attribute.
        new._scale = self._scale * (unit // self._unit)
        # Return the new instance.
        return new

    def _subscript_standard(self, indices):
        """Perform standard array subscription.

        This method handles cases involving slices, an ellipsis, or integers,
        including v[:], v[...], v[i, :], v[:, j], and v[i, j], where i and j are
        integers.
        """
        result = self._get_array(indices)
        Type = Scalar if isinstance(result, numbers.Number) else Array
        return Type(result, unit=self.unit, name=self.name)

    def _subscript_custom(self, args):
        """Perform array subscription specific to this object.

        This method handles all cases that don't meet the criteria for
        `_subscript_standard`.
        """
        if not isinstance(args, (tuple, list)):
            args = [args]
        expanded = self._expand_ellipsis(args)
        idx = [
            range(self.shape[i])
            if isinstance(arg, slice) else arg
            for i, arg in enumerate(expanded)
        ]
        indices = numpy.ix_(*list(idx))
        return self._copy_with(data=self._get_array(indices))

    def _expand_ellipsis(
        self,
        user: typing.Sequence,
    ) -> typing.Tuple[slice, ...]:
        """Expand an ``Ellipsis`` into one or more ``slice`` objects."""
        if all(idx is not Ellipsis for idx in user):
            return user
        length = self.ndim - len(user) + 1
        start = user.index(Ellipsis)
        return (
            *user[slice(start)],
            *([slice(None)] * length),
            *user[slice(start+length, self.ndim)],
        )

    _HANDLED_TYPES = (numpy.ndarray, numbers.Number, list)

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        """Provide support for `numpy` universal functions.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information on use of this special method.

        See
        https://numpy.org/doc/stable/reference/generated/numpy.lib.mixins.NDArrayOperatorsMixin.html#numpy.lib.mixins.NDArrayOperatorsMixin
        for the specific implementation example that this class follows.

        Notes
        -----
        This method first ensures that the input types (as well as the type of
        `out`, if passed via keyword) are supported types. It then extracts
        appropriate operands from any arguments that are an instance of this
        class or a subclass and applies `ufunc`. If `ufunc` is one of a
        pre-defined set of methods that would cause an attribute of this class
        to become ambiguous or undefined, this method return the result of
        applying `ufunc`. Otherwise, this method will return a new instance with
        zero or more attributes that have been modified in a `ufunc`-specific
        way (e.g., multiplying two instances' units when `ufunc` is 'multiply').
        The default behavior is therefore to return a new instance with
        unmodified attributes.
        """
        out = kwargs.get('out', ())
        for x in args + out:
            if not isinstance(x, self._HANDLED_TYPES + (Array,)):
                return NotImplemented
        if out:
            kwargs['out'] = tuple(
                x._array if isinstance(x, Array)
                else x for x in out
            )
        name = ufunc.__name__
        operands = self._ufunc_hook(ufunc, *args)
        compute = getattr(ufunc, method)
        data = compute(*operands, **kwargs)
        evaluate = (
            self.meta[name].evaluate if name in self.meta
            else self.meta.implement(compute)
        )
        kwds = {k: v for k, v in kwargs.items() if k != 'out'}
        meta = evaluate(*args, **kwds)
        if type(data) is tuple:
            return tuple(
                self._new_from_func(x, updates=meta)
                for x in data
            )
        if method == 'at':
            return None
        return self._new_from_func(data, updates=meta)

    def _ufunc_hook(self, ufunc, *inputs):
        """Convert input object types based on `ufunc`."""
        if ufunc.__name__ == 'power':
            inputs = [
                float(x) if isinstance(x, numbers.Real)
                else x for x in inputs
            ]
        return tuple(
            x._array if isinstance(x, Array)
            else x for x in inputs
        )

    _HANDLED_FUNCTIONS = {}

    def __array_function__(self, func, types, args, kwargs):
        """Provide support for functions in the `numpy` public API.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for more
        information of use of this special method. The implementation shown here
        is a combination of the example on that page and code from the
        definition of `EncapsulateNDArray.__array_function__` in
        https://github.com/dask/dask/blob/main/dask/array/tests/test_dispatch.py

        The initial `issubclass` check allows subclasses that don't override
        `__array_function__` to handle objects of this type.
        """
        accepted = (Array, numpy.ndarray, numpy.ScalarType)
        if not all(issubclass(ti, accepted) for ti in types):
            return NotImplemented
        if func in self._HANDLED_FUNCTIONS:
            result = self._HANDLED_FUNCTIONS[func](*args, **kwargs)
            return self._new_from_func(result)
        args = tuple(
            arg._array if isinstance(arg, Array)
            else arg for arg in args
        )
        types = tuple(
            ti for ti in types
            if not issubclass(ti, Array)
        )
        data = self._array
        arr = data.__array_function__(func, types, args, kwargs)
        return self._new_from_func(arr)

    def _new_from_func(self, result, updates: dict=None):
        """Create a new instance from the result of a `numpy` function.
        
        If `result` is a `numpy.ndarray` and `updates` is a (possibly empty)
        `dict`, this method will create a new instance. Otherwise, it will
        return `result` as-is.
        """
        if isinstance(result, numpy.ndarray) and isinstance(updates, dict):
            return self._copy_with(data=result, **updates)
        return result

    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Support casting to `numpy` array types.
        
        Notes
        -----
        This will retrieve the underlying array before applying `*args` and
        `**kwargs`, in order to avoid a `TypeError` when using
        `netCDF4.Dataset`. See
        https://github.com/mcgibbon/python-examples/blob/master/scripts/file-io/load_netCDF4_full.py
        """
        data = self._array
        return numpy.asanyarray(data, *args, **kwargs)

    @property
    def _array(self):
        """The full data array for internal use."""
        return self._get_array()

    def _get_array(self, index: IndexLike=None):
        """Load and scale the array data corresponding to `index`.
        
        Notes
        -----
        - This method represents the internal interface to the underlying array.
        - General users should subscript instances via the standard bracket
          syntax.
        - This method defers the process of loading the appropriate array
          without any numerical scaling to `self._load_array()`, then rescales
          the result if the unit has changed.
        """
        return self._scale * self._load_array(index)

    def _load_array(self, index=None):
        """Get the unscaled array from disk or memory.

        If `index` is ``None`` or an empty iterable, this method will produce
        the entire array. Otherwise, it will create the requested subarray from
        `self.data`. It will always attempt to use a cached version of the full
        array before loading from disk.
        
        The reasoning behind this algorithm is as follows: If we need to load
        the full array at any point, we may as well save it because subscripting
        an in-memory `numpy.ndarray` is much faster than re-reading from disk
        for large arrays. However, we should avoid reading in the full array if
        the caller only wants a small portion of it. We don't cache these
        subarrays because reusing a subarray is only meaningful if the indices
        haven't changed. Furthermore, accessing a subarray via bracket syntax
        creates a new object, at which point the subarray becomes the new
        object's full array.
        """
        if iterables.missing(index):
            if self._full_array is not None:
                return self._full_array
            array = self._read_array()
            self._full_array = array
            return array
        if self._full_array is not None:
            idx = numpy.index_exp[index]
            return self._full_array[idx]
        return self._read_array(index)

    def _read_array(self, index: IndexLike=None):
        """Read the array data from disk.
        
        If `index` is "missing" in the sense defined by `~iterables.missing`
        this method will load and return the full array. If `index` is not
        missing, this method will first attempt to subscript `self.data` before
        converting it to an array and returning it. If it catches either a
        `TypeError` or an `IndexError`, it will create the full array before
        subscripting and returning it. The former may occur if `self.data` is a
        sequence type like `list`, `tuple`, or `range`; the latter may occur
        when attempting to subscript certain array-like objects (e.g.,
        `netCDF4._netCDF4.Variable`) with valid `numpy` index expressions.
        """
        if not iterables.missing(index):
            idx = numpy.index_exp[index]
            try:
                return numpy.asarray(self.data[idx])
            except (TypeError, IndexError):
                return numpy.asarray(self.data)[idx]
        return numpy.asarray(self.data)

    def _copy_with(self, **updates):
        """Create a new instance from the current attributes."""
        data = updates.get('data', self.data)
        meta = {
            k: updates.get(k, getattr(self, k))
            for k in self.meta.parameters
        }
        return type(self)(data, **meta)

    @classmethod
    def implements(cls, numpy_function):
        """Register an `__array_function__` implementation for this class.

        See https://numpy.org/doc/stable/reference/arrays.classes.html for the
        suggestion on which this method is based.

        EXAMPLE
        -------
        Overload `numpy.mean` for an existing class called ``Array`` with a
        version that accepts no keyword arguments::

            @Array.implements(numpy.mean)
            def mean(a: Array, **kwargs) -> Array:
                if kwargs:
                    msg = "Cannot pass keywords to numpy.mean with Array"
                    raise TypeError(msg)
                return numpy.sum(a) / len(a)

        This will compute the mean of the underlying data when called with no
        arguments, but will raise an exception when called with arguments:

            >>> v = Array([[1, 2], [3, 4]])
            >>> numpy.mean(v)
            5.0
            >>> numpy.mean(v, axis=0)
            ...
            TypeError: Cannot pass keywords to numpy.mean with Array
        """
        def decorator(func):
            cls._HANDLED_FUNCTIONS[numpy_function] = func
            return func
        return decorator





import collections
import collections.abc
import contextlib
import numbers
import operator as standard
import typing

import numpy

from goats.core import aliased
from goats.core import axis
from goats.core import computed
from goats.core import constant
from goats.core import iterables
from goats.core import metadata
from goats.core import metric
from goats.core import observed
from goats.core import physical
from goats.core import reference
from goats.core import symbolic
from goats.core import variable


class Parameters(collections.abc.Sequence, iterables.ReprStrMixin):
    """The parameters of an `~observing.Quantity`."""

    def __init__(self, *names: str) -> None:
        self._names = self._init(*names)

    def _init(self, *args):
        names = iterables.unique(*iterables.unwrap(args, wrap=tuple))
        if not names or all(isinstance(name, str) for name in names):
            return names
        raise TypeError(
            f"Can't initialize instance of {type(self)}"
            f" with {names!r}"
        )

    __abs__ = metadata.identity(abs)
    """Called for abs(self)."""
    __pos__ = metadata.identity(standard.pos)
    """Called for +self."""
    __neg__ = metadata.identity(standard.neg)
    """Called for -self."""

    __add__ = metadata.identity(standard.add)
    """Called for self + other."""
    __sub__ = metadata.identity(standard.sub)
    """Called for self - other."""

    def merge(a, *others):
        """Return the unique axis names in order."""
        names = list(a._names)
        for b in others:
            if isinstance(b, Parameters):
                names.extend(b._names)
        return Parameters(*set(names))

    __mul__ = merge
    """Called for self * other."""
    __rmul__ = merge
    """Called for other * self."""
    __truediv__ = merge
    """Called for self / other."""

    def __pow__(self, other):
        """Called for self ** other."""
        if isinstance(other, numbers.Real):
            return self
        return NotImplemented

    def __eq__(self, other):
        """True if self and other represent the same axes."""
        return (
            isinstance(other, Parameters) and other._names == self._names
            or (
                isinstance(other, str)
                and len(self) == 1
                and other == self._names[0]
            )
            or (
                isinstance(other, typing.Iterable)
                and len(other) == len(self)
                and all(i in self for i in other)
            )
        )

    def __hash__(self):
        """Support use as a mapping key."""
        return hash(self._names)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._names)

    def __getitem__(self, __i: typing.SupportsIndex):
        """Called for index-based access."""
        return self._names[__i]

    def __str__(self) -> str:
        return f"[{', '.join(repr(name) for name in self._names)}]"


class Quantity(variable.Quantity):
    """A quantity with one or more name(s), a unit, and axes."""

    _parameters: Parameters=None

    def __init__(self, __data, **meta) -> None:
        super().__init__(__data, **meta)
        parsed = self.parse_attrs(__data, meta, parameters=())
        self._parameters = Parameters(parsed['parameters'])
        self.meta.register('parameters')
        self.display.register('parameters')
        self.display['__str__'].append("parameters={parameters}")
        self.display['__repr__'].append("parameters={parameters}")

    def parse_attrs(self, this, meta: dict, **targets):
        if (
            isinstance(this, variable.Quantity)
            and not isinstance(this, Quantity)
        ): # barf
            meta.update({k: getattr(this, k) for k in ('unit', 'name', 'axes')})
            this = this.data
        return super().parse_attrs(this, meta, **targets)

    @property
    def parameters(self):
        """The optional parameters that define this observing quantity."""
        return Parameters(self._parameters)


class Interface(collections.abc.Collection):
    """An interface to quantities required to make observations."""

    def __init__(
        self,
        axes: axis.Interface,
        variables: variable.Interface,
        constants: constant.Interface,
    ) -> None:
        self.axes = axes
        """The axis-managing objects available to this observer."""
        self.variables = variables
        """The variable quantities available to this observer."""
        self.constants = constants
        """The constant quantities available to this observer."""
        self._system = None
        self._names = None
        self._functions = None
        assumptions = {
            k: v
            for k, v in constants.items(aliased=True)
            if isinstance(v, constant.Assumption)
        } if constants else {}
        self.assumptions = aliased.Mapping(assumptions)
        """The default physical assumptions."""

    def __contains__(self, __x) -> bool:
        return __x in self.names

    def __iter__(self) -> typing.Iterator:
        return iter(self.names)

    def __len__(self) -> int:
        return len(self.names)

    @property
    def system(self):
        """The metric system of this observer's dataset quantities."""
        if self._system is None:
            self._system = self.variables.system
        return self._system

    @property
    def names(self):
        """The names of all available quantities."""
        if self._names is None:
            self._names = list(self.variables) + list(self.functions)
        return self._names

    @property
    def functions(self):
        """The computable quantities available to this observer."""
        if self._functions is None:
            self._functions = computed.Interface(self.axes, self.variables)
        return self._functions

    def get_observable(self, key: str):
        """Get the observable quantity corresponding to `key`."""
        for this in (self.variables, self.functions):
            if key in this:
                return this[key]

    # TODO: Refactor `get_unit` and `get_axes` to reduce overlap.

    def get_unit(self, key: str):
        """Determine the unit corresponding to `key`."""
        if key in self.variables:
            return self.variables[key].unit
        if key in self.functions:
            return self.functions[key].unit
        s = str(key)
        expression = symbolic.Expression(reference.NAMES.get(s, s))
        term = expression[0]
        this = self.get_unit(term.base) ** term.exponent
        if len(expression) > 1:
            for term in expression[1:]:
                this *= self.get_unit(term.base) ** term.exponent
        return metadata.Unit(this)

    def get_axes(self, key: str):
        """Determine the axes corresponding to `key`."""
        if key in self.variables:
            return self.variables[key].axes
        if key in self.functions:
            return self.functions[key].axes
        s = str(key)
        expression = symbolic.Expression(reference.NAMES.get(s, s))
        term = expression[0]
        this = self.get_axes(term.base)
        if len(expression) > 1:
            for term in expression[1:]:
                this *= self.get_axes(term.base)
        return metadata.Axes(this)

    def compute_index(self, key: str, **constraints) -> axis.Index:
        """Compute the axis-indexing object for `key`."""
        if key not in self.axes:
            raise ValueError(f"No axis corresponding to {key!r}") from None
        if key not in constraints:
            return self.axes[key].index()
        this = constraints[key]
        if isinstance(this, axis.Index):
            return this
        return self.axes[key].index(*iterables.whole(this))

    def compute_value(self, key: str, **constraints) -> physical.Scalar:
        """Create a parameter value for `key`."""
        if key not in self.constants and key not in constraints:
            raise ValueError(
                f"No parameter corresponding to {key!r}"
            ) from None
        if key in constraints:
            return self._compute_value(constraints[key])
        return self._compute_value(self.constants[key])

    def _compute_value(self, this):
        """Compute a parameter value."""
        # TODO: Generalize beyond scalar parameters.
        scalar = physical.scalar(this)
        unit = self.system.get_unit(unit=scalar.unit)
        return scalar[unit]


class Application:
    """A general observing application."""

    def __init__(self, interface: Interface) -> None:
        self.interface = interface
        """The interface to an observer's dataset."""
        self.user = None
        """The current set of user constraints."""
        self._coordinates = None
        self._cache = {}

    def __contains__(self, __k: str):
        """True if the named constraint affects this application."""
        # Should this check all dimensions and parameters or only user
        # constraints?
        return __k in self.user

    def apply(self, **constraints):
        """Update the observing constraints.
        
        This method will apply the given constraints when computing axis indices
        and parameter values within the context of this application. It will
        also clear the corresponding items from the instance cache.
        """
        if self.user is None:
            self.user = {}
        for key in constraints:
            for subset in self._cache.values():
                if key in subset:
                    del subset[key]
        self.user.update(constraints)
        return self

    def reset(self, *keys: str):
        """Reset constraints to their default values.
        
        Parameters
        ----------
        *keys : string
            Zero or more names of constraints to reset. This method will also
            remove all named items from the instance cache By default, this
            method will reset all constraints and clear the cache.
        """
        if not keys:
            self.user = {}
            self._cache = {}
        else:
            subsets = self._cache.values()
            for key in keys:
                if key in self.user:
                    del self.user[key]
                for subset in subsets:
                    if key in subset:
                        del subset[key]
        return self

    def observe(self, key: str) -> Quantity:
        """Create an observation within the context of this application."""
        expression = symbolic.Expression(reference.NAMES.get(key, key))
        term = expression[0]
        result = self.get_observable(term.base)
        if len(expression) == 1:
            # We don't need to multiply or divide quantities.
            if term.exponent == 1:
                # We don't even need to raise this quantity to a power.
                return result
            return result ** term.exponent
        q0 = result ** term.exponent
        if len(expression) > 1:
            for term in expression[1:]:
                result = self.get_observable(term.base)
                q0 *= result ** term.exponent
        return q0

    def evaluate(self, q) -> Quantity:
        """Create an observing quantity based on the given quantity."""
        if isinstance(q, computed.Quantity):
            parameters = [
                parameter for parameter in q.parameters
                if parameter in self.interface.constants
            ]
            return Quantity(self.compute(q), parameters=parameters)
        if isinstance(q, variable.Quantity):
            return Quantity(self.process(q))
        raise ValueError(f"Unknown quantity: {q!r}") from None

    def process(self, q: variable.Quantity) -> variable.Quantity:
        """Compute observer-specific updates to a variable quantity."""
        if any(alias in self.coordinates for alias in q.name):
            # This is an axis-reference quantity.
            return self._subscript(q)
        needed = self._compute_coordinates(q)
        if not needed:
            # There are no axes over which to interpolate.
            return self._subscript(q)
        return self._interpolate(q)

    def compute(self, q: computed.Quantity) -> variable.Quantity:
        """Determine dependencies and compute the result of this function."""
        dependencies = {p: self.get_dependency(p) for p in q.parameters}
        return q(**dependencies)

    def get_dependency(self, key: str):
        """Get the named constant or variable quantity."""
        if this := self.get_observable(key):
            return this
        return self.get_value(key)

    def get_observable(self, key: str):
        """Retrieve and evaluate an observable quantity."""
        if quantity := self.interface.get_observable(key):
            return self.evaluate(quantity)

    def _subscript(self, q: variable.Quantity, *axes: str):
        """Extract a subset of this quantity."""
        if not axes:
            return q[tuple(self.get_index(a) for a in q.axes)]
        indices = [
            self.get_index(a) if a in axes else slice(None)
            for a in q.axes
        ]
        return q[tuple(indices)]

    def _compute_coordinates(self, q: variable.Quantity):
        """Determine the measurable observing indices."""
        dimensions = self._compute_dimensions(q)
        coordinates = {}
        if not dimensions:
            return coordinates
        for a in q.axes:
            idx = self.get_index(a)
            if a in dimensions and idx.unit is not None:
                coordinates[a] = {
                    'targets': numpy.array(idx.values),
                    'reference': self.coordinates[a],
                }
        return coordinates

    def _compute_dimensions(self, q: variable.Quantity):
        """Determine over which axes to interpolate, if any."""
        coordinates = {}
        references = []
        for a in q.axes:
            idx = self.get_index(a)
            if idx and idx.unit is not None:
                coordinates[a] = idx.values
                references.append(self.coordinates[a])
        return [
            a for (a, c), r in zip(coordinates.items(), references)
            if not numpy.all([r.array_contains(target) for target in c])
        ]

    def _interpolate(self, q: variable.Quantity):
        """Interpolate the given quantity within this context."""
        raise NotImplementedError

    @property
    def coordinates(self):
        """The reference quantities for dataset coordinate axes."""
        if self._coordinates is None:
            coordinates = self._build_coordinates()
            self._coordinates = aliased.Mapping(coordinates)
        return self._coordinates

    def _build_coordinates(self):
        """Helper for `~coordinates` property. Extracted for overloading."""
        return {
            k: self.interface.variables.get(k)
            for k in self.interface.axes.keys(aliased=True)
        }

    def get_index(self, key: str) -> axis.Index:
        """Get the axis-indexing object for `key`."""
        if 'indices' not in self._cache:
            self._cache['indices'] = {}
        if key in self._cache['indices']:
            return self._cache['indices'][key]
        with contextlib.suppress(ValueError):
            idx = self.interface.compute_index(key, **self.user)
            self._cache['indices'][key] = idx
            return idx

    def get_value(self, key: str) -> physical.Scalar:
        """Get the parameter value correpsonding to `key`."""
        if 'values' not in self._cache:
            self._cache['values'] = {}
        if key in self._cache['values']:
            return self._cache['values'][key]
        with contextlib.suppress(ValueError):
            val = self.interface.compute_value(key, **self.user)
            self._cache['values'][key] = val
            return val


class Implementation:
    """The base observing implementation."""

    def __init__(
        self,
        name: str,
        interface: Interface,
    ) -> None:
        """
        Initialize this instance.

        Parameters
        ----------
        name : string
            The name of the quantity to observe.

        interface : `~Interface`
            The interface to all observing-related quantities.
        """
        self.name = name
        self.interface = interface
        self._cache = {}

    # TODO:
    # - refactor `get_unit` and `get_axes` to reduce overlap.
    # - redefine `_cache` as an aliased mapping.

    def get_unit(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Return the unit corresponding to `name`."""
        if 'unit' not in self._cache:
            self._cache['unit'] = {}
        if name in self._cache['unit']:
            return self._cache['unit'][name]
        unit = self._lookup_unit(name)
        self._cache['unit'][name] = unit
        return unit

    def _lookup_unit(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Determine the unit corresponding to `name`."""
        if isinstance(name, str):
            return self.interface.get_unit(name)
        unit = (self.interface.get_unit(key) for key in name)
        return next(unit, None)

    def get_axes(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Axes:
        """Return the axes corresponding to `name`."""
        if 'axes' not in self._cache:
            self._cache['axes'] = {}
        if name in self._cache['axes']:
            return self._cache['axes'][name]
        axes = self._lookup_axes(name)
        self._cache['axes'][name] = axes
        return axes

    def _lookup_axes(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> metadata.Unit:
        """Determine the axes corresponding to `name`."""
        if isinstance(name, str):
            return self.interface.get_axes(name)
        axes = (self.interface.get_axes(key) for key in name)
        return next(axes, None)

    def apply(
        self,
        application: Application,
        unit: metadata.UnitLike=None,
    ) -> observed.Quantity:
        """Apply an observing context to the target quantity."""
        result = application.observe(self.name)
        indices = {k: application.get_index(k) for k in result.axes}
        scalars = {k: application.get_value(k) for k in result.parameters}
        context = observed.Context(indices, scalars=scalars)
        if unit:
            return observed.Quantity(result[unit], context)
        return observed.Quantity(result, context)



import pathlib
import typing
import numbers

import numpy
import numpy.typing

from goats import Environment
from ..core import (
    axis,
    aliased,
    datafile,
    fundamental,
    iterables,
    iotools,
    measurable,
    metadata,
    metric,
    numerical,
    observer,
    observing,
    physical,
    reference,
    variable,
)
from . import interpolation
from .runtime import BaseTypesH


ENV = Environment('eprem')


basetypes = BaseTypesH(source=ENV['src'])


class Axes(axis.Interface):
    """Interface to EPREM axis-indexing objects."""

    def __init__(
        self,
        data: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        super().__init__(data, system)
        self._time = None
        self._shell = None
        self._species = None
        self._energy = None
        self._mu = None
        defined = {
            'time': self.time,
            'shell': self.shell,
            'species': self.species,
            'energy': self.energy,
            'mu': self.mu,
        }
        self.update(defined)

    @property
    def shell(self):
        """Indexer for the EPREM shell dimension."""
        if self._shell is None:
            def method(targets, unit):
                # NOTE: The presence of `unit` is a hack because 'shell'
                # currently gets a unit of '1' even though it should probably be
                # None. This hack is due to the design of `axis.Quantity`.
                return axis.Data(targets)
            self._shell = axis.Indexer(method, len(self.variables['shell']))
        return self._shell

    @property
    def time(self):
        """Indexer for the EPREM time dimension."""
        if self._time is None:
            this = self.variables['time']
            self._time = axis.Indexer(self._build_coordinate(this), len(this))
        return self._time

    @property
    def mu(self):
        """Indexer for the EPREM pitch-angle dimension."""
        if self._mu is None:
            this = self.variables['mu']
            self._mu = axis.Indexer(self._build_coordinate(this), len(this))
        return self._mu

    @property
    def energy(self):
        """Indexer for the EPREM energy dimension."""
        if self._energy is None:
            this = self.variables['energy']
            def method(
                targets,
                unit: metadata.UnitLike,
                species: typing.Union[str, int]=0,
            ) -> axis.Data:
                s = self.species.compute([species]).points
                t = (
                    numpy.squeeze(targets[s, :])
                    if getattr(targets, 'ndim', None) == 2
                    else targets
                )
                compute = self._build_coordinate(numpy.squeeze(this[s, :]))
                return compute(t, unit)
            self._energy = axis.Indexer(method, this.shape[1])
        return self._energy

    @property
    def species(self):
        """Indexer for the EPREM species dimension."""
        if self._species is None:
            mass = self.variables['mass']['nuc']
            charge = self.variables['charge']['e']
            symbols = fundamental.elements(mass, charge)
            def method(targets):
                indices = [
                    symbols.index(target)
                    if isinstance(target, str) else int(target)
                    for target in targets
                ]
                return axis.Data(indices, values=symbols)
            self._species = axis.Indexer(method, len(symbols))
        return self._species

    def _build_coordinate(self, this: variable.Quantity):
        """Create coordinate-like axis data from the given variable."""
        def method(targets, unit: typing.Union[str, metadata.Unit]):
            if not targets:
                # If there are no target values, we assume the user wants the
                # entire axis.
                return axis.Data(range(len(this)), values=numpy.array(this))
            if all(isinstance(t, typing.SupportsIndex) for t in targets):
                # All the target values are already indices.
                return axis.Data(targets, values=numpy.array(this))
            measured = measurable.measure(targets)
            if measured.unit != '1':
                # If the targets include a dimensioned unit, we want to
                # initialize the array with that unit.
                array = physical.Array(measured.values, unit=measured.unit)
            else:
                # If the measured unit is dimensionless, it could be because the
                # targets truly are dimensionless or because the user wants to
                # use the default unit. Since we have no choice but to assume
                # that the calling object (probably an instance of
                # `core.axis.Quantity`) passed an appropriate default unit,
                # which may be dimensionless, the default unit is the
                # appropriate unit for both cases.
                array = physical.Array(measured.values, unit=unit)
            # Convert the reference variable quantity to the default unit.
            converted = this[unit]
            if array.unit | converted.unit: # Could also use try/except
                array = array[converted.unit]
            values = numpy.array(array)
            indices = [
                numerical.find_nearest(this, float(value)).index
                for value in values
            ]
            return axis.Data(indices, values=values)
        return method


class Application(observing.Application):
    """The EPREM-specific observing context."""

    # TODO:
    # - generalize interpolation from this subpackage
    # - implement generalized interpolation in core.observing.Application
    # - add EPREM-specific interpolation here
    def process(self, q: variable.Quantity) -> variable.Quantity:
        """Compute EPREM-specific updates to a variable quantity."""
        if any(alias in self.coordinates for alias in q.name):
            # This is an axis-reference quantity.
            return self._subscript(q)
        needed = self._compute_coordinates(q)
        if not needed:
            # There are no axes over which to interpolate.
            return self._subscript(q)
        new = self._interpolate(q, needed.values())
        # We only want to subscript the uninterpolated axes.
        interpolated = [
            'shell' if d == 'radius' else d
            for d in needed
        ]
        axes = list(set(q.axes) - set(interpolated))
        return self._subscript(new, *axes)

    def _build_coordinates(self):
        base = super()._build_coordinates()
        grid = {
            (k, *self.interface.variables.alias(k, include=True)):
            self.interface.variables[k]
            for k in {'radius', 'theta', 'phi'}
        }
        return {**base, **grid}

    _axes = {
        'time': 0,
        'energy': 1,
        'mu': 1,
    }

    def _compute_coordinates(self, q: variable.Quantity):
        base = {
            k: {**c, 'axis': self._axes.get(k)}
            for k, c in super()._compute_coordinates(q).items()
        }
        for key in reference.ALIASES['radius']:
            if values := self.get_value(key):
                try:
                    iter(values)
                except TypeError:
                    floats = [float(values)]
                else:
                    floats = [float(value) for value in values]
                base['radius'] = {
                    'targets': numpy.array(floats),
                    'reference': self.coordinates['radius'],
                    'axis': 1,
                }
        return base

    def _compute_dimensions(self, q: variable.Quantity):
        base = super()._compute_dimensions(q)
        if (
            'shell' in q.axes
            and any(r in self for r in reference.ALIASES['radius'])
        ): base.append('radius')
        return base

    def _interpolate(
        self,
        q: variable.Quantity,
        coordinates: typing.Iterable[typing.Dict[str, typing.Any]],
    ) -> variable.Quantity:
        """Internal interpolation logic."""
        array = None
        for coordinate in coordinates:
            array = self._interpolate_coordinate(
                q,
                coordinate['targets'],
                coordinate['reference'],
                axis=coordinate.get('axis'),
                workspace=array,
            )
        meta = {k: getattr(q, k, None) for k in {'unit', 'name', 'axes'}}
        return variable.Quantity(array, **meta)

    def _interpolate_coordinate(
        self,
        q: variable.Quantity,
        targets: numpy.ndarray,
        reference: variable.Quantity,
        axis: int=None,
        workspace: numpy.ndarray=None,
    ) -> numpy.ndarray:
        """Interpolate a variable array based on a known coordinate."""
        array = numpy.array(q) if workspace is None else workspace
        indices = (q.axes.index(d) for d in reference.axes)
        dst, src = zip(*enumerate(indices))
        reordered = numpy.moveaxis(array, src, dst)
        interpolated = interpolation.apply(
            reordered,
            numpy.array(reference),
            targets,
            axis=axis,
        )
        return numpy.moveaxis(interpolated, dst, src)


Instance = typing.TypeVar('Instance', bound='Observer')


class Observer(observer.Interface, iterables.ReprStrMixin):
    """Base class for EPREM observers."""

    _templates: typing.Iterable[typing.Callable] = None

    def __init__(
        self,
        __id: int,
        source: iotools.PathLike=ENV['source'],
        config: iotools.PathLike=ENV['config'],
        system: str='mks',
    ) -> None:
        self._id = __id
        super().__init__(
            'preEruption', 'phiOffset',
            system=system,
            apply=Application,
        )
        self._confpath = None
        self._dataset = None
        self.readfrom(source, config=config)

    def readfrom(
        self: Instance,
        source: iotools.PathLike,
        config: iotools.PathLike=None
    ) -> Instance:
        # Use arguments to update paths.
        datapath = self._build_datapath(source)
        confpath = self._build_confpath(config, directory=datapath.parent)
        # Update the internal data interface.
        dataset = datafile.Interface(datapath)
        axes = Axes(dataset, system=self.system())
        variables = variable.Interface(dataset, system=self.system())
        constants = runtime.Arguments(
            source_path=ENV['src'],
            config_path=confpath,
        )
        self._data = observing.Interface(axes, variables, constants)
        # Update other attributes if everything succeeded.
        self._dataset = dataset
        self._confpath = confpath
        return super().readfrom(datapath)

    def _build_datapath(self, directory: iotools.PathLike):
        """Create the path to the dataset from `directory`."""
        default = ENV['source'] or pathlib.Path.cwd()
        this = iotools.ReadOnlyPath(directory or default)
        if this.is_dir():
            path = iotools.find_file_by_template(
                self._templates,
                self._id,
                directory=this,
            )
            return iotools.ReadOnlyPath(path)
        raise TypeError(
            f"Can't create path to dataset from {directory!r}"
        ) from None

    def _build_confpath(
        self,
        config: iotools.PathLike,
        directory: iotools.PathLike=pathlib.Path.cwd(),
    ) -> iotools.ReadOnlyPath:
        """Create the path to the configuration file."""
        if not config: # use default directory and name
            return iotools.ReadOnlyPath(directory / ENV['config'])
        this = pathlib.Path(config)
        if this.is_dir(): # use default directory
            return iotools.ReadOnlyPath(this / ENV['config'])
        if this.name == config: # use default name
            return iotools.ReadOnlyPath(directory / this)
        raise ValueError(
            f"Can't create path to configuration file from {config!r}"
        ) from None

    @property
    def confpath(self) -> iotools.ReadOnlyPath:
        """The full path to this observer's runtime parameter file."""
        return self._confpath

    @property
    def datapath(self) -> iotools.ReadOnlyPath:
        """The path to this observer's dataset."""
        return self._source

    @property
    def dataset(self) -> datafile.Interface:
        """This observer's original dataset."""
        return self._dataset

    def __str__(self) -> str:
        return str(self.datapath)


class Stream(Observer):
    """An EPREM stream observer."""

    _templates = [
        lambda n: f'obs{n:06}.nc',
        lambda n: f'flux{n:06}.nc',
    ]


class Point(Observer):
    """An EPREM point observer."""

    _templates = [
        lambda n: f'p_obs{n:06}.nc',
    ]




from goats.core import physical
from goats.core import iterables
from goats.core import fundamental


class Species(iterables.ReprStrMixin):
    """A single species in a plasma."""

    def __init__(
        self,
        symbol: str=None,
        mass: float=None,
        charge: float=None,
    ) -> None:
        self._symbol = symbol
        self._mass = mass
        self._charge = charge
        if self._symbol is None and self._mass is None and self._charge is None:
            raise ValueError("Element is undefined")

    @property
    def symbol(self) -> str:
        """The elemental symbol of this species."""
        if self._symbol is None:
            s = fundamental.elements([self._mass], [self._charge])
            self._symbol = s[0]
        return self._symbol

    @property
    def mass(self) -> physical.Scalar:
        """The mass of this species."""
        if self._mass is None:
            base = self._symbol.rstrip('+-')
            element = fundamental.ELEMENTS.find(base, unique=True)
            value = element['mass']
            unit = 'nucleon'
            aliases = ['m', 'mass']
            self._mass = physical.Scalar(value, unit=unit, name=aliases)
        return self._mass

    @property
    def m(self):
        """Alias for mass."""
        return self.mass

    @property
    def charge(self) -> physical.Scalar:
        """The charge of this species."""
        if self._charge is None:
            base = self._symbol.rstrip('+-')
            sign = self._symbol.lstrip(base)
            value = sum(float(f"{s}1.0") for s in sign)
            unit = 'e'
            aliases = ['q', 'charge']
            self._charge = physical.Scalar(value, unit=unit, name=aliases)
        return self._charge

    @property
    def q(self):
        """Alias for charge."""
        return self.charge

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return f"{self.symbol!r}, mass={self.m}, charge={self.q}"





import collections
import numbers
import typing

from goats.core import aliased
from goats.core import datafile
from goats.core import iterables
from goats.core import metadata
from goats.core import metric
from goats.core import reference
from goats.core import variable


T = typing.TypeVar('T')


class Data(iterables.ReprStrMixin):
    """Index points and corresponding values."""

    def __new__(cls, points, **kwargs):
        if not all(isinstance(i, numbers.Integral) for i in points):
            raise ValueError(
                "All index points must have integral type"
            ) from None
        return super().__new__(cls)

    def __init__(
        self,
        points: typing.Iterable[numbers.Integral],
        values: typing.Iterable[typing.Union[numbers.Real, str]]=None,
    ) -> None:
        self.points = tuple(points)
        """The integral index points."""
        self.values = self.points if iterables.missing(values) else values
        """The values associated with index points."""

    def __str__(self) -> str:
        return ', '.join(str((i, j)) for i, j in zip(self.points, self.values))


class Index(iterables.ReprStrMixin, collections.UserList):
    """A sequence of indices representing axis values."""

    def __init__(
        self,
        data: Data,
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None:
        super().__init__(list(data.points))
        self._values = data.values
        self._unit = metadata.Unit(unit) if unit is not None else None
        self._name = name

    @property
    def values(self):
        """The axis value at each index."""
        return self._values

    @property
    def unit(self):
        """The metric unit of the corresponding values, if any."""
        return self._unit

    @property
    def name(self):
        """The name of the axis that these indices represent."""
        return self._name

    def __str__(self) -> str:
        """A simplified representation of this object."""
        string = f"{str(self.name)!r}, values={self.values}"
        if self.unit:
            string += f", unit={str(self.unit)!r}"
        return string


class Indexer(iterables.ReprStrMixin):
    """An object that computes axis indices from user values."""

    def __init__(
        self,
        method: typing.Callable[..., Data],
        size: int,
    ) -> None:
        self._method = method
        """The method that converts target values into indices."""
        self.size = size
        """The maximum number of indices."""

    def compute(self, *args, **kwargs):
        """Call the index-computing method."""
        return self._method(*args, **kwargs)

    def normalize(self, *user: T):
        """Convert user input into suitable target values."""
        if not user:
            return range(self.size)
        if isinstance(user[0], slice):
            return iterables.slice_to_range(user[0], stop=self.size)
        if isinstance(user[0], range):
            return user[0]
        return user

    def __str__(self) -> str:
        return f"{self._method.__qualname__}, size={self.size}"


Instance = typing.TypeVar('Instance', bound='Quantity')


class Quantity(metadata.NameMixin, iterables.ReprStrMixin):
    """A callable representation of a single dataset axis."""

    @typing.overload
    def __init__(
        self: Instance,
        __indexer: Indexer,
        *,
        unit: metadata.UnitLike=None,
        name: typing.Union[str, typing.Iterable[str]]=None,
    ) -> None:
        """Create a new axis from scratch."""

    @typing.overload
    def __init__(
        self: Instance,
        instance: Instance,
    ) -> None:
        """Create a new axis from an existing instance."""

    def __init__(self, __a, **meta) -> None:
        try:
            self._unit = meta.pop('unit')
        except KeyError:
            self._unit = None
        self._indexer, kwd = self.parse_attrs(__a, meta, name='')
        self._name = metadata.Name(kwd['name'])

    def parse_attrs(
        self,
        this: typing.Union['Quantity', Indexer],
        meta: dict,
        **targets
    ) -> typing.Tuple[Indexer, dict]:
        """Get instance attributes from initialization arguments."""
        if isinstance(this, Quantity):
            return this._indexer, {k: getattr(this, k) for k in targets}
        return this, {k: meta.get(k, v) for k, v in targets.items()}

    @property
    def unit(self):
        if self._unit is not None:
            return metadata.Unit(self._unit)

    def index(self, *args, **kwargs):
        """Convert arguments into an index-like quantity."""
        targets = self._indexer.normalize(*args)
        if self.unit is None:
            # This axis does not represent a measurable quantity, so the
            # index-computing method should not expect a unit and the resulting
            # index object will not have a unit.
            return Index(
                self._indexer.compute(targets, **kwargs),
                name=self.name,
            )
        # This axis represents a measurable quantity, so we want to make sure
        # the index-computing method and the resulting index object use
        # consistent units.
        unit = kwargs.pop('unit', self.unit)
        return Index(
            self._indexer.compute(targets, unit, **kwargs),
            unit=unit,
            name=self.name
        )

    def __str__(self) -> str:
        """A simplified representation of this object."""
        string = f"{str(self.name)!r}"
        if self.unit:
            string += f", unit={str(self.unit)!r}"
        return string


class IndexTypeError(Exception):
    """Invalid index argument."""


class Interface(aliased.MutableMapping):
    """An interface to the array axes available from a dataset."""

    def __init__(
        self,
        dataset: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        self.dataset = dataset
        """The dataset to which these axes belong."""
        self._variables = None
        self._system = metric.System(system or 'mks')
        indexers = {
            k: self.build_default(k)
            for k in self.dataset.axes.keys(aliased=True)
        }
        super().__init__(indexers)

    @property
    def variables(self):
        """The variable quantities that these axes support."""
        if self._variables is None:
            self._variables = variable.Interface(
                self.dataset,
                system=self.system,
            )
        return self._variables

    @property
    def system(self):
        """The metric system of these axes."""
        return self._system

    def __getitem__(self, __k: str) -> Quantity:
        """Get the named axis object, if possible."""
        unit = self.get_unit(__k)
        name = self.get_name(__k)
        try:
            return Quantity(
                self._get_indexer(__k),
                unit=unit,
                name=name,
            )
        except KeyError as err:
            raise KeyError(
                f"No known indexing method for {__k}"
            ) from err

    def _get_indexer(self, key: str):
        """Get the axis indexer for `key`, or use the default."""
        if key in self.keys():
            return super().__getitem__(key)
        raise KeyError(f"No indexing method for axis {key!r}")

    def build_default(self, key: str):
        """Define the default axis-indexer method."""
        n = self.dataset.axes[key].size
        def method(targets):
            try:
                indices = [int(arg) for arg in targets]
                if all(0 <= idx < n for idx in indices):
                    return Data(indices)
            except TypeError as err:
                raise IndexTypeError(
                    f"Can't convert {targets!r} to integer indices."
                ) from err
            raise ValueError(
                f"One or more index {targets} is outside the interval"
                " [0, {n-1}]"
            ) from None
        return Indexer(method, n)

    def get_unit(self, key: str):
        """Get the metric unit corresponding to `key`."""
        try:
            unit = self.variables[key].unit
        except KeyError:
            unit = None
        return unit

    def get_name(self, key: str):
        """Get the set of aliases for `key`."""
        return reference.NAMES.get(key, key)

    def resolve(
        self,
        names: typing.Iterable[str],
        mode: str='strict',
    ) -> typing.Tuple[str]:
        """Compute and order the available axes in `names`."""
        axes = self.dataset.available('axes').canonical
        ordered = tuple(name for name in axes if name in names)
        if mode == 'strict':
            return ordered
        extra = tuple(name for name in names if name not in ordered)
        if not extra:
            return ordered
        if mode == 'append':
            return ordered + extra
        raise ValueError(f"Unrecognized mode {mode!r}")






"""
Objects and functions for operating on metadata.
"""

import abc
import collections.abc
import fractions
import inspect
import math
import numbers
import operator as standard
import typing

import numpy

from goats.core import aliased
from goats.core import iterables
from goats.core import metric
from goats.core import utilities


T = typing.TypeVar('T')


def consistent(name: str, *args):
    """Check consistency of a metadata attribute across `args`."""
    values = [
        utilities.getattrval(arg, name) for arg in args
        if hasattr(arg, name)
    ]
    if not values:
        return True # trivial case
    v0 = values[0]
    return all([v == v0 for v in values])


class NTypesError(Exception):
    """Inconsistent number of types in a new rule."""


class Operands(collections.abc.Sequence, iterables.ReprStrMixin):
    """A sequence of operands."""

    def __init__(self, *args: T, reference: T=None) -> None:
        self._args = list(args)
        self._reference = reference
        self._types = None

    @property
    def reference(self):
        """The reference object.
        
        Notes
        -----
        - The default reference object is the first argument used to initialize
          this instance.
        - The reference object is not part of the sequence of operands and
          therefore will not explicitly appear when iterating over an instance.
        """
        if self._reference is None:
            self._reference = self[0]
        return self._reference

    def agree(self, *names: str, strict: bool=False):
        """Compare values of named attribute(s) across operands.
        
        This method determines if all the named attributes have the same value
        when present in an operand. The result is trivially true for a single
        operand
        """
        if len(self) == 1:
            return True
        if strict:
            for name in names:
                values = self.getall(name, mode='strict')
                v0 = values[0]
                if any(value != v0 for value in values):
                    return False
            return True
        targets = self.find(*names)
        for name in names:
            values = [
                utilities.getattrval(target, name)
                for target in targets
            ]
            v0 = values[0]
            if any(value != v0 for value in values):
                return False
        return True

    def find(self, *names: str):
        """Get all operands with the named attributes."""
        return [
            obj for obj in self
            if all(hasattr(obj, name) for name in names)
        ]

    def allhave(self, *names: str):
        """Determine if all operands have the named attribute(s)."""
        return all(hasattr(obj, name) for obj in self for name in names)

    def getany(self, name: str):
        """Get the a value for the named attribute, if possible."""
        valid = (v for v in self.getall(name, mode='next') if v)
        return next(valid, None)

    def getall(self, name: str, mode: str=None):
        """Get operand values for the named attribute."""
        return [self._get(name, operand, mode) for operand in self]

    def _get(self, name: str, operand, mode: str=None):
        """Helper for `~Operands.getall`."""
        available = hasattr(operand, name)
        if mode == 'next':
            return getattr(operand, name) if available else None
        if not available and mode == 'strict':
            raise AttributeError(f"{operand!r} has no attribute {name!r}")
        if available or mode == 'force':
            return utilities.getattrval(operand, name)
        return utilities.getattrval(self.reference, name)

    @property
    def types(self):
        """The type of each object."""
        if self._types is None:
            self._types = tuple(type(i) for i in self)
        return self._types

    @typing.overload
    def __getitem__(self, __i: typing.SupportsIndex) -> T: ...

    @typing.overload
    def __getitem__(self, __s: slice) -> 'Operands': ...

    def __getitem__(self, __i):
        """Access operands by index."""
        if isinstance(__i, typing.SupportsIndex):
            return self._args[__i]
        return Operands(*self._args[__i])

    def __len__(self) -> int:
        """The number of operands. Called for len(self)."""
        return len(self._args)

    def __iter__(self):
        """Iterate over operands. Called for iter(self)."""
        yield from self._args

    def __eq__(self, other):
        """Called for self == other."""
        if not len(self) == len(other):
            return False
        return all(i in other for i in self)

    def __str__(self) -> str:
        return ', '.join(str(i) for i in self)


class Operator(iterables.ReprStrMixin):
    """A generalized metadata operator."""

    def __init__(self, __callable: typing.Callable) -> None:
        self.method = __callable

    def __call__(self, name: str, operands: Operands, **kwargs):
        """Apply this operator's method to the given operands."""
        values = operands.getall(name, mode='force')
        try:
            return self.method(*values, **kwargs)
        except TypeError as err:
            if len(operands) == 1:
                return values[0]
            if operands.allhave(name):
                raise MetadataError(err) from err
            raise err

    def __str__(self) -> str:
        return str(self.method)


class Operation(typing.Generic[T], iterables.ReprStrMixin):
    """A general operation context."""

    def __init__(
        self,
        __type: typing.Type[T],
        *parameters: str,
        name: str=None,
    ) -> None:
        self._type = __type
        self._parameters = list(parameters)
        self.name = name
        """The name of this operation, if any."""
        self.supported = set()
        """The type(s) of operand(s) for which this operation is valid."""
        self.suppressed = set()
        """The type(s) of operand(s) for which this operation is invalid."""
        self.nargs = None
        """The number of required arguments in this operation."""

    @property
    def parameters(self):
        """The names of metadata attributes."""
        return tuple(self._parameters)

    @property
    def implemented(self):
        """False if this operation supports no operand types."""
        return bool(self.supported)

    def copy(self):
        """Create a deep copy of this operation."""
        return Operation(
            self._type,
            *self.parameters,
            name=self.name,
        )

    def update(self, *parameters: str):
        """Update the names of metadata attributes."""
        if not parameters:
            raise ValueError(
                "Update requires at least one parameter."
            ) from None
        self._parameters.extend(parameters)
        return self

    Types = typing.Union[
        type,
        typing.Tuple[type],
        typing.List[type],
    ]

    def suppress(self, *types: Types, symmetric: bool=False):
        """Suppress operations on these types.
        
        Parameters
        ----------
        *types
            One or more type specifications to suppress. A type specification
            consists of one or more type(s) of operand(s) that are
            inter-operable. Multiple type specifications must be grouped into
            lists or tuples, even if they represent a single type; this method
            will interpret ungrouped types as a single type specification.

        symmetric : bool, default=False
            If true, suppress the forward and reverse version of each type
            specification. Otherwise, suppress only the given type
            specifications. This option can simplify method calls for binary
            operations; it has no effect on unary operations.
        """
        self.suppressed.add(types)
        if symmetric:
            self.suppressed.add(types[::-1])
        return self

    def support(self, *types: Types, symmetric: bool=False):
        """Support operations on these types.
        
        Parameters
        ----------
        *types
            One or more type specifications to support. A type specification
            consists of one or more type(s) of operand(s) that are
            inter-operable. Multiple type specifications must be grouped into
            lists or tuples, even if they represent a single type; this method
            will interpret ungrouped types as a single type specification.

        symmetric : bool, default=False
            If true, support the forward and reverse version of each type
            specification. Otherwise, support only the given type
            specifications. This option can simplify method calls for binary
            operations; it has no effect on unary operations.

        Raises
        ------
        NTypesError
            The number of types in a given type specification is not equal to
            the number of types in existing type specifications.
        """
        if not types:
            raise ValueError("No types to add.") from None
        these = [types] if all(isinstance(t, type) for t in types) else types
        for this in these:
            ntypes = len(this)
            if self.nargs is None:
                self.nargs = ntypes
            if ntypes != self.nargs:
                raise NTypesError(
                    f"Can't add a length-{ntypes} rule to a collection"
                    f" of length-{self.nargs} rules."
                ) from None
            self.supported.add(this)
            if symmetric:
                self.supported.add(this[::-1])
        return self

    def supports(self, *types: Types):
        """Determine if this operation supports `types` or subtypes."""
        if self.nargs and len(types) != self.nargs:
            return False
        if types in self.suppressed:
            return False
        for t in self.suppressed:
            if all(issubclass(i, j) for i, j in zip(types, t)):
                return False
        if types in self.supported:
            return True
        for t in self.supported:
            if all(issubclass(i, j) for i, j in zip(types, t)):
                return True
        return self._type in types

    def evaluate(self, *args, **kwargs):
        """Compute attribute values using the default method."""
        return self.operator(*args, **kwargs)

    @property
    def operator(self):
        """The operator created from this operation's default callable."""
        return self.apply(REFERENCE[self.name])

    def apply(self, method: typing.Callable):
        """Create an operator by applying the given method.
        
        Parameters
        ----------
        method : callable
            The callable object that should operate on metadata attributes.

        Notes
        -----
        * If the method is not callable, this method assumes that there is no
          metadata to compute. This is the case, for example, with type casts
          and binary comparisons.
        """
        compute = Operator(method)
        def operator(*args, **kwargs):
            # NOTE: A more conservative approach would be to raise an error for
            # un-callable methods and force downstream code to define a trivial
            # function that returns `None` for any arguments.
            if not callable(method):
                return None
            operands = Operands(*args)
            types = operands.types
            if not self.supports(*types):
                raise OperandTypeError(
                    f"Can't apply {method.__qualname__!r} to metadata"
                    f" with types {', '.join(t.__qualname__ for t in types)}"
                ) from None
            return {p: compute(p, operands, **kwargs) for p in self.parameters}
        operator.__name__ = self.name or f'__{method.__name__}__'
        operator.__doc__ = method.__doc__
        if callable(method) and not isinstance(method, type):
            operator.__text_signature__ = str(inspect.signature(method))
        return operator

    def __eq__(self, __o):
        """Determine if two operational contexts are equal."""
        return (
            isinstance(__o, Operation)
            and __o._type == self._type
            and __o.supported == self.supported
            and __o.suppressed == self.suppressed
        )


class OperandTypeError(Exception):
    """Operands are incompatible for a given operation."""


class MetadataError(Exception):
    """Error in computing metadata value."""


class OperatorFactory(collections.abc.Mapping):
    """A factory for objects that operate on metadata."""

    def __init__(self, __type: T, *parameters: str) -> None:
        """
        Initialize this instance.

        Parameters
        ----------
        __type : type
            The type of object to which these operators will apply.

        *parameters : string
            Zero or more strings representing the updatable attributes in each
            operand to these operations.
        """
        self._type = __type
        self._parameters = list(iterables.unique(*parameters))
        self._operations = None

    @property
    def parameters(self):
        """The names of updatable metadata attributes."""
        return tuple(self._parameters)

    def register(self, *names: str):
        """Register additional names of metadata attributes."""
        self._parameters.extend(iterables.unique(*names))
        for operation in self.operations.values():
            operation.update(*names)
        return self

    def implement(self, method: typing.Callable):
        """Apply `method` to the default implementation."""
        return Operation(self._type, *self.parameters).apply(method)

    def check(self, *args):
        """Ensure that all arguments have consistent metadata values."""
        for p in self.parameters:
            if not consistent(p, *args):
                raise ValueError(f"Inconsistent metadata for {p!r}")

    def consistent(self, *args):
        """True if all arguments have consistent metadata values."""
        try:
            self.check(*args)
        except ValueError:
            return False
        else:
            return True

    def __getitem__(self, __k: str):
        """Retrieve the appropriate operation context."""
        if __k in self.operations:
            return self.operations[__k]
        raise KeyError(f"Unknown context {__k!r}") from None

    def __len__(self) -> int:
        """The number of defined operations."""
        return len(self.operations)

    def __iter__(self):
        """Iterate over operation contexts."""
        return iter(self.operations)

    @property
    def operations(self) -> typing.Mapping[str, Operation]:
        """The operations defined here.
        
        Notes
        -----
        Operations are not the same as operators. For example, the special
        operator methods `__truediv__`, `__rtruediv__`, and `__itruediv__`, as
        well as the `numpy` ufunc `true_divide`, all correspond to the 'true
        divide' operation. This aliased mapping of operations does not currently
        support look-up by operator name in order to avoid misleading the user
        into thinking there are different objects for different operators
        corresponding to the same operation.
        """
        if self._operations is None:
            operations = aliased.MutableMapping.fromkeys(REFERENCE)
            for k in REFERENCE:
                operations[k] = Operation(
                    self._type,
                    *self.parameters,
                    name=k,
                )
            self._operations = operations
        return self._operations


def identity(__operator: typing.Callable):
    """Create an operator that immediately returns its argument."""
    def operator(*args: T):
        first = args[0]
        for arg in args:
            if type(arg) == type(first) and arg != first:
                return NotImplemented
        return first
    operator.__name__ = f'__{__operator.__name__}__'
    operator.__doc__ = __operator.__doc__
    return operator


def suppress(__operator: typing.Callable):
    """Unconditionally suppress an operation."""
    def operator(*args, **kwargs):
        return NotImplemented
    operator.__name__ = f'__{__operator.__name__}__'
    operator.__doc__ = __operator.__doc__
    return operator


def restrict(__f: typing.Callable, *types: type, reverse: bool=False):
    """Restrict allowed operand types for an operation."""
    s = 'r' if reverse else ''
    name = f"__{s}{__f.__name__}__"
    def operator(a, b, **kwargs):
        method = getattr(super(type(a), a), name)
        if not isinstance(b, (type(a), *types)):
            return NotImplemented
        return method(b, **kwargs)
    operator.__name__ = name
    operator.__doc__ = __f.__doc__
    return operator


class UnitError(Exception):
    """Base class for Unit-related exceptions."""


class DimensionMismatch(UnitError):
    """These units have different dimensions."""


class ScaleMismatch(UnitError):
    """These units have different metric scale factors."""


class UnitLike(metaclass=abc.ABCMeta):
    """The definition of a unit-like metadata attribute.
    
    All concrete and virtual subclasses of this class can serve as a unit
    attribute and downstream code may use this class for instance checks.
    """

UnitLike.register(str)
UnitLike.register(metric.Unit)


_metadata_mixins = (
    numpy.lib.mixins.NDArrayOperatorsMixin,
    iterables.ReprStrMixin,
)


class Attribute(*_metadata_mixins):
    """Base class for metadata attributes."""

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        """Provide support for `numpy` universal functions."""
        if func := getattr(self, f'_ufunc_{ufunc.__name__}', None):
            return func(*args, **kwargs)
        return NotImplemented

    def __ne__(self, __o) -> bool:
        return not __o == self


class Unit(metric.Unit, Attribute):
    """The unit attribute of a quantity."""

    def _ufunc_sqrt(self, arg):
        """Implement the square-root function for a unit."""
        return arg**0.5

    def __mul__(self, other):
        """Called for self * other."""
        if isinstance(other, UnitLike):
            return super().__mul__(other)
        return self

    def __truediv__(self, other):
        """Called for self / other."""
        if isinstance(other, UnitLike):
            return super().__truediv__(other)
        return self

    __rmul__ = __mul__
    """Called for other * self."""

    def __rtruediv__(self, other):
        """Called for other / self."""
        if isinstance(other, UnitLike):
            return super().__rtruediv__(other)
        return self

    # NOTE: This class implements in-place multiplication and division with the
    # following trivial definitions because otherwise, control will pass to
    # `Attribute.__array_ufunc__`, which will (correctly) fail to find
    # `_ufunc_multiply` or `_ufunc_divide`, result in an exception.

    __imul__ = __mul__
    """Called for self *= other."""

    __itruediv__ = __truediv__
    """Called for self /= other"""

    def __add__(self, other):
        """Called for self + other; either a no-op or an error."""
        return self._add_sub(other)

    def __sub__(self, other):
        """Called for self - other; either a no-op or an error."""
        return self._add_sub(other)

    def _add_sub(self, other):
        """Called for self +/- other; either a no-op or an error."""
        if not isinstance(other, UnitLike):
            return NotImplemented
        if self == other:
            return self
        if isinstance(other, Unit):
            errmsg = "The units '{}' and '{}' have different {}"
            if self.dimensions == other.dimensions:
                raise ScaleMismatch(
                    errmsg.format(self, other, 'scale factors')
                ) from None
            raise DimensionMismatch(
                errmsg.format(self, other, 'dimensions')
            ) from None
        raise UnitError(
            f"The units '{self}' and '{other}' are incompatible"
        ) from None


class Name(collections.abc.Collection, Attribute):
    """The name attribute of a quantity."""

    def __init__(self, *aliases: str) -> None:
        self._aliases = aliased.MappingKey(*aliases)

    def add(self, aliases: typing.Union[str, typing.Iterable[str]]):
        """Add `aliases` to this name."""
        self._aliases = self._aliases | aliases
        return self

    def _ufunc_sqrt(self, arg):
        """Implement the square-root function for a name."""
        return arg**'1/2'

    def _implement(symbol: str, reverse: bool=False, strict: bool=False):
        """Implement a symbolic operation.

        This function creates a new function that symbolically represents the
        result of applying an operator to two operands, `a` and `b`.
        
        Parameters
        ----------
        symbol : string
            The string representation of the operator.

        reverse : bool, default=False
            If true, apply the operation to reflected operands.

        strict : bool, default=False
            If true, require that `b` be an instance of `a`'s type or a subtype.

        Raises
        ------
        TypeError
            `strict` is true and `b` is not an instance of `a`'s type or a
            subtype.

        Notes
        -----
        * The nullspace for names is the empty string.
        * 'a | A' + 2 -> undefined
        * 'a | A' + 'a | A' -> 'a | A' when `strict` is true
        * 'a | A' * 'a | A' -> 'a*a | A*A' when `strict` is false
        * 'a | A' + 'b | B' -> 'a+b | a+B | A+b | A+B'
        * 'a | A' * 'b | B' -> 'a*b | a*B | A*b | A*B'
        * 'a | A' * 2 -> 'a*2 | A*2'
        """
        def compute(a, b):
            """Symbolically combine `a` and `b`."""
            x, y = (b, a) if reverse else (a, b)
            if isinstance(y, typing.Iterable) and not isinstance(y, str):
                return Name([f'{i}{symbol}{j}' for i in x for j in y])
            try:
                fixed = fractions.Fraction(y)
            except ValueError:
                fixed = y
            t = '{1}{s}{0}' if reverse else '{0}{s}{1}'
            return Name([t.format(i, fixed, s=symbol) for i in x])
        def operator(self, that):
            if not self or not that:
                return Name([''])
            if strict:
                if not isinstance(that, type(self)):
                    raise TypeError(
                        f"Can't apply {symbol} "
                        f"to {type(self)!r} and {type(that)!r}"
                    ) from None
                if that == self:
                    return self
            if that == self:
                return Name([f'{i}{symbol}{i}' for i in self])
            if str(that) == '1':
                return Name([str(i) for i in self])
            return compute(that, self) if reverse else compute(self, that)
        s = f"other {symbol} self" if reverse else f"self {symbol} other"
        operator.__doc__ = f"Called for {s}"
        return operator

    __add__ = _implement(' + ', strict=True)
    __radd__ = _implement(' + ', strict=True, reverse=True)
    __sub__ = _implement(' - ', strict=True)
    __rsub__ = _implement(' - ', strict=True, reverse=True)
    __mul__ = _implement(' * ')
    __rmul__ = _implement(' * ', reverse=True)
    __truediv__ = _implement(' / ')
    __rtruediv__ = _implement(' / ', reverse=True)
    __pow__ = _implement('^')
    __rpow__ = _implement('^', reverse=True)

    def __pos__(self):
        """Called for +self."""
        if not self:
            return Name()
        return Name([f"+{i}" for i in self])

    def __neg__(self):
        """Called for -self."""
        if not self:
            return Name()
        return Name([f"-{i}" for i in self])

    def __abs__(self):
        """Called for abs(self)."""
        if not self:
            return Name()
        return Name([f"abs({i})" for i in self])

    def __round__(self):
        """Called for round(self)."""
        if not self:
            return Name()
        return Name([f"round({i})" for i in self])

    def __floor__(self):
        """Called for math.floor(self)."""
        if not self:
            return Name()
        return Name([f"floor({i})" for i in self])

    def __ceil__(self):
        """Called for math.ceil(self)."""
        if not self:
            return Name()
        return Name([f"ceil({i})" for i in self])

    def __trunc__(self):
        """Called for math.trunc(self)."""
        if not self:
            return Name()
        return Name([f"trunc({i})" for i in self])

    def __bool__(self) -> bool:
        return bool(self._aliases)

    def __hash__(self) -> int:
        return hash(self._aliases)

    def __contains__(self, __x) -> bool:
        return __x in self._aliases

    def __iter__(self) -> typing.Iterator:
        return iter(self._aliases)

    def __len__(self) -> int:
        return len(self._aliases)

    def __eq__(self, __o) -> bool:
        if isinstance(__o, Name):
            return __o._aliases == self._aliases
        try:
            return __o == self._aliases
        except TypeError:
            return False

    def __str__(self) -> str:
        return str(self._aliases)


class Axes(collections.abc.Sequence, iterables.ReprStrMixin):
    """A representation of one or more axis names."""

    def __init__(self, *names: str) -> None:
        self._names = self._init(*names)

    def _init(self, *args):
        names = iterables.unwrap(args, wrap=tuple)
        if all(isinstance(name, str) for name in names):
            return names
        raise TypeError(
            f"Can't initialize instance of {type(self)}"
            f" with {names!r}"
        )

    __abs__ = identity(abs)
    """Called for abs(self)."""
    __pos__ = identity(standard.pos)
    """Called for +self."""
    __neg__ = identity(standard.neg)
    """Called for -self."""

    __add__ = identity(standard.add)
    """Called for self + other."""
    __sub__ = identity(standard.sub)
    """Called for self - other."""

    def merge(a, *others):
        """Return the unique axis names in order."""
        names = list(a._names)
        for b in others:
            if isinstance(b, Axes):
                names.extend(b._names)
        return Axes(*iterables.unique(*names))

    __mul__ = merge
    """Called for self * other."""
    __rmul__ = merge
    """Called for other * self."""
    __truediv__ = merge
    """Called for self / other."""

    def __pow__(self, other):
        """Called for self ** other."""
        if isinstance(other, numbers.Real):
            return self
        return NotImplemented

    def __eq__(self, other):
        """True if self and other represent the same axes."""
        return (
            isinstance(other, Axes) and other._names == self._names
            or (
                isinstance(other, str)
                and len(self) == 1
                and other == self._names[0]
            )
            or (
                isinstance(other, typing.Iterable)
                and len(other) == len(self)
                and all(i in self for i in other)
            )
        )

    def __hash__(self):
        """Support use as a mapping key."""
        return hash(self._names)

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self._names)

    def __getitem__(self, __i: typing.SupportsIndex):
        """Called for index-based access."""
        return self._names[__i]

    def __str__(self) -> str:
        return f"[{', '.join(repr(name) for name in self._names)}]"


class UnitMixin:
    """Mixin class for quantities with a unit."""

    _unit: Unit=None

    @property
    def unit(self):
        """This quantity's metric unit."""
        return Unit(self._unit)

    def convert(self, unit: UnitLike):
        """Set the unit of this object's values."""
        if unit != self._unit:
            new = Unit(unit)
            self.apply_conversion(new)
            self._unit = new
        return self

    def apply_conversion(self, new: Unit):
        """Update data values for unit conversion.
        
        Classes that use this mixin class to manage a unit attribute may
        overload this method to customize the result of updating the unit (e.g.,
        scaling a data attribute). The default implementation does nothing.
        """
        pass


class NameMixin:
    """Mixin class for quantities with a name."""

    _name: Name=None

    @property
    def name(self):
        """This quantity's name."""
        return Name(self._name)

    def alias(self, *updates: str, reset: bool=False):
        """Set or add to this object's name(s)."""
        aliases = updates if reset else self.name.add(updates)
        self._name = Name(*aliases)
        return self


class AxesMixin:
    """Mixin class for quantities with axes."""

    _axes: Axes=None

    @property
    def axes(self):
        """This quantity's indexable axes."""
        return Axes(self._axes)


# def base(*names: str, result: str=None):
#     """Create a base class for handling metadata attributes."""

#     known = {
#         'unit': {'type': Unit, 'default': '1', 'mixin': UnitMixin},
#         'name': {'type': Name, 'default': '', 'mixin': NameMixin},
#         'axes': {'type': Axes, 'default': (), 'mixin': AxesMixin},
#     }
#     for name in names:
#         if name not in known:
#             raise ValueError(f"Unknown metadata attribute: {name!r}") from None

#     class Base:
#         """A base class for quantities with metadata attributes."""

#         def __init__(self, **meta) -> None:
#             self.meta = OperatorFactory(type(self))
#             for name in names:
#                 this = known[name]
#                 value = this['type'](meta.get(name, this['default']))
#                 setattr(self, f"_{name}", value)
#                 self.meta.register(name)

#     mixins = [known[name]['mixin'] for name in names if name in known]
#     return type(result or 'Interface', (Base, *mixins), {})


# Measurable = base('unit', result='Measurable')
# """An object with a unit."""


# Identifiable = base('name', result='Identifiable')
# """An object with a name."""


# Locatable = base('axes', result='Locatable')
# """An object with axes."""


# Distinguishable = base('unit', 'name', result='Distinguishable')
# """A measurable and identifiable object."""


# Observable = base('unit', 'name', 'axes', result='Observable')
# """A distinguishable and locatable object."""


# ATTRIBUTES = {
#     'unit': {'type': Unit, 'default': '1', 'mixin': UnitMixin},
#     'name': {'type': Name, 'default': '', 'mixin': NameMixin},
#     'axes': {'type': Axes, 'default': (), 'mixin': AxesMixin},
# }


# Class = typing.TypeVar('Class', bound=type)


# class Subtype(typing.Generic[T]):
#     """Generic type to indicate a subclass in type hints."""


# def includes(*names: str):
#     """Decorator for adding metadata attributes to a class."""
#     for name in names:
#         if name not in ATTRIBUTES:
#             raise ValueError(
#                 f"Unknown metadata attribute: {name!r}"
#             ) from None
#     mixins = [
#         ATTRIBUTES[name]['mixin']
#         for name in names if name in ATTRIBUTES
#     ]
#     _names = [f"_{name}" for name in names]
#     def decorator(cls: typing.Type[T]):
#         class Interface:
#             """A class with dynamically generated metadata attributes."""
#             def __init__(self) -> None:
#                 self.meta = OperatorFactory(type(self))
#                 for name, _name in zip(names, _names):
#                     user = getattr(self, _name)
#                     attribute = ATTRIBUTES[name]
#                     value = attribute['type'](user or attribute['default'])
#                     setattr(self, _name, value)
#                     self.meta.register(name)
#         bases = [cls, Interface, *mixins]
#         return type(cls.__name__, tuple(bases), {k: None for k in _names})
#     return decorator


# class Interface:
#     """"""

#     def __new__(cls: Class, *args, **attributes) -> Class:
#         names = list(attributes)
#         mixins = [
#             ATTRIBUTES[name]['mixin']
#             for name in names if name in ATTRIBUTES
#         ]
#         namespace = {f'_{name}': None for name in names}
#         return type(cls.__name__, (cls, *mixins), namespace)

#     def __init__(self, **attributes) -> None:
#         self.meta=OperatorFactory(type(self))
#         for name, given in attributes.items():
#             _name = f"_{name}"
#             attribute = ATTRIBUTES[name]
#             value = attribute['type'](given or attribute['default'])
#             setattr(self, _name, value)
#             self.meta.register(name)


# def interface(*names: str):
#     """"""
#     for name in names:
#         if name not in ATTRIBUTES:
#             raise ValueError(
#                 f"Unknown metadata attribute: {name!r}"
#             ) from None
#     mixins = [
#         ATTRIBUTES[name]['mixin']
#         for name in names if name in ATTRIBUTES
#     ]
#     _names = [f"_{name}" for name in names]
#     class Metadata:
#         """A class with dynamically generated metadata attributes."""
#         def __init__(self, **attributes) -> None:
#             self.meta=OperatorFactory(type(self))
#             for name, given in attributes.items():
#                 _name = f"_{name}"
#                 attribute = ATTRIBUTES[name]
#                 value = attribute['type'](given or attribute['default'])
#                 setattr(self, _name, value)
#                 self.meta.register(name)
#     bases = (Metadata, *mixins)
#     return type('Metadata', bases, {k: None for k in _names})


class Interface:
    """Support for metadata attributes."""

    def __init__(self, **attributes) -> None:
        pass


_reference: typing.Dict[str, dict] = {
    'int': {
        'callable': None,
    },
    'float': {
        'callable': None,
    },
    'abs': {
        'callable': abs,
        'aliases': ['absolute'],
    },
    'pos': {
        'callable': standard.pos,
        'aliases': ['positive'],
    },
    'neg': {
        'callable': standard.neg,
        'aliases': ['negative'],
    },
    'ceil': {
        'callable': math.ceil,
        'aliases': ['ceiling'],
    },
    'floor': {
        'callable': math.floor,
    },
    'trunc': {
        'callable': math.trunc,
        'aliases': ['truncate'],
    },
    'round': {
        'callable': round,
    },
    'lt': {
        'callable': None,
        'aliases': ['less'],
    },
    'le': {
        'callable': None,
        'aliases': ['less_equal', 'less equal'],
    },
    'gt': {
        'callable': None,
        'aliases': ['greater'],
    },
    'ge': {
        'callable': None,
        'aliases': ['greater_equal', 'greater equal'],
    },
    'add': {
        'callable': standard.add,
    },
    'sub': {
        'callable': standard.sub,
        'aliases': ['subtract'],
    },
    'mul': {
        'callable': standard.mul,
        'aliases': ['multiply'],
    },
    'truediv': {
        'callable': standard.truediv,
        'aliases': ['true_divide', 'true divide'],
    },
    'pow': {
        'callable': pow,
        'aliases': ['power'],
    },
}
REFERENCE = aliased.Mapping(_reference).squeeze(strict=True)





import typing

import numpy

from goats.core import aliased
from goats.core import axis
from goats.core import iterables
from goats.core import constant
from goats.core import physical
from goats.core import variable


class Context:
    """The context of an observation."""

    def __init__(
        self,
        indices: typing.Mapping[str, axis.Index],
        scalars: typing.Mapping[str, physical.Scalar]=None
    ) -> None:
        self._indices = indices
        self._scalars = scalars
        self._assumptions = None
        self._axes = None

    @property
    def axes(self):
        """The arrays of axis values relevant to the observation."""
        if self._axes is None:
            items = (
                self._indices.items(aliased=True)
                if isinstance(self._indices, aliased.Mapping)
                else self._indices.items()
            )
            axes = {
                k: physical.Array(
                    index.values,
                    unit=index.unit,
                    name=index.name,
                )
                for k, index in items
                # HACK: This handles cases in which a function added a
                # non-standard axis. For example, integral flux removes the
                # 'energy' axis and adds a 'minimum energy' axis. A long-term
                # solution should treat non-standard axes as equivalent to
                # standard axes.
                if index
            }
            self._axes = aliased.Mapping(axes)
        return self._axes

    @property
    def assumptions(self):
        """The physical assumptions relevant to this observation."""
        if self._assumptions is None:
            items = (
                self._scalars.items(aliased=True)
                if isinstance(self._scalars, aliased.Mapping)
                else self._scalars.items()
            )
            assumptions = {
                k: constant.Assumption(scalar)
                for k, scalar in items if scalar
            }
            self._assumptions = aliased.Mapping(assumptions)
        return self._assumptions

    def __eq__(self, __o) -> bool:
        """True if two contexts have equal axes and assumptions."""
        if isinstance(__o, Context):
            return all(
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('axes', 'assumptions')
            )
        return NotImplemented


class Quantity(iterables.ReprStrMixin):
    """The result of observing an observable quantity."""

    def __init__(
        self,
        __v: variable.Quantity,
        context: Context,
    ) -> None:
        self._quantity = __v
        self._context = context
        self._array = None
        self._data = None
        self._parameters = None

    @property
    def array(self):
        """The observed data array, with singular dimensions removed.
        
        This property intends to provide a convenient shortcut for cases in
        which the observation result is effectively N-dimensional but singular
        dimensions cause it to appear to have higher dimensionality.
        """
        if self._array is None:
            self._array = numpy.squeeze(self.data)
        return self._array

    @property
    def data(self):
        """The observed variable quantity.
        
        This property provides direct access to the array interface, as well as
        to metadata properties of the observed quantity.
        """
        if self._data is None:
            self._data = self._quantity
        return self._data

    @property
    def parameters(self):
        """The names of scalar assumptions relevant to this observation."""
        if self._parameters is None:
            self._parameters = list(self._context.assumptions)
        return self._parameters

    def __getitem__(self, __x):
        """Get a scalar by name or array values by index."""
        if __x in self._context.axes:
            return self._context.axes[__x]
        if __x in self._context.assumptions:
            return self._context.assumptions[__x]

    def __eq__(self, __o) -> bool:
        """True if two instances have equivalent attributes."""
        if isinstance(__o, Quantity):
            return all(
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('data', 'context')
            )
        return NotImplemented

    def __str__(self) -> str:
        """A simplified representation of this object."""
        attrs = [
            f"unit='{self.data.unit}'",
            f"dimensions={self.data.axes}",
            f"parameters={self.parameters}",
        ]
        return f"'{self.data.name}': {', '.join(attrs)}"




"""
Tools for reading data and metadata from file.
"""

import abc
import collections.abc
import typing

import netCDF4
import numpy
import numpy.typing

from goats.core import aliased
from goats.core import iotools
from goats.core import iterables
from goats.core import reference


class Viewer(collections.abc.Mapping):
    """An abstract base class for data-viewing objects."""

    def __init__(self, path: iotools.ReadOnlyPath) -> None:
        self.members = self.get_members(path)

    def __iter__(self) -> typing.Iterator:
        return iter(self.members)

    def __len__(self) -> int:
        return len(self.members)

    @abc.abstractmethod
    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        """Get the appropriate members for this viewer."""
        pass

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return '\n\n'.join(f"{k}:\n{v!r}" for k, v in self.items())

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        return f"\n::{self.__class__.__qualname__}::\n\n{self}"


class Variable(typing.NamedTuple):
    """A dataset variable."""

    data: numpy.typing.ArrayLike
    unit: str
    axes: typing.Tuple[str]
    name: str

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return self._display(sep='\n')

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        display = self._display(sep='\n', tab=4)
        return f"{self.__class__.__qualname__}(\n{display}\n)"

    def _display(
        self,
        sep: str=', ',
        tab: int=0,
    ) -> str:
        """Helper for `__str__` and `__repr__`."""
        attrs = [
            f"data={type(self.data)}",
            f"unit={self.unit!r}",
            f"axes={self.axes}",
            f"name={self.name!r}",
        ]
        indent = ' ' * tab
        return sep.join(f"{indent}{attr}" for attr in attrs)


class NetCDFVariables(Viewer):
    """An object for viewing variables in a NetCDF dataset."""

    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        dataset = netCDF4.Dataset(path, 'r')
        return dataset.variables

    def __getitem__(self, name: str) -> Variable:
        if name in self.members:
            data = self.members[name]
            unit = self._get_unit_from_data(data)
            axes = self._get_axes_from_data(data)
            return Variable(data, unit, axes, name)
        raise KeyError(f"No variable called '{name}'")

    def _get_axes_from_data(self, data):
        """Compute appropriate variable axes from a dataset object."""
        return tuple(getattr(data, 'dimensions', ()))

    def _get_unit_from_data(self, data):
        """Compute appropriate variable units from a dataset object."""
        available = (
            getattr(data, attr) for attr in ('unit', 'units')
            if hasattr(data, attr)
        )
        return next(available, None)


class Axis(typing.NamedTuple):
    """A dataset axis."""

    size: int
    name: str

    def __str__(self) -> str:
        """A simplified representation of this object."""
        return f"{self.name!r}, size={self.size}"

    def __repr__(self) -> str:
        """An unambiguous representation of this object."""
        module = f"{self.__module__.replace('goats.', '')}."
        name = self.__class__.__qualname__
        return f"{module}{name}({self})"


class NetCDFAxes(Viewer):
    """An object for viewing axes in a NetCDF dataset."""

    def get_members(self, path: iotools.ReadOnlyPath) -> typing.Mapping:
        dataset = netCDF4.Dataset(path, 'r')
        return dataset.dimensions

    def __getitem__(self, name: str) -> Axis:
        if name in self.members:
            data = self.members[name]
            size = getattr(data, 'size', None)
            return Axis(size, name)
        raise KeyError(f"No axis corresponding to {name!r}")


class ViewerFactory(collections.abc.MutableMapping):
    """A class that creates appropriate viewers for a dataset."""

    _viewer_map = {
        '.nc': {
            'variables': NetCDFVariables,
            'axes': NetCDFAxes,
        }
    }

    def __init__(self, path: iotools.ReadOnlyPath) -> None:
        self._viewers = self._get_viewers(path)
        self.path = path

    def __iter__(self) -> typing.Iterator:
        return iter(self._viewers)

    def __len__(self) -> int:
        return len(self._viewers)

    def _get_viewers(
        self,
        path: iotools.ReadOnlyPath,
    ) -> typing.Dict[str, typing.Type[Viewer]]:
        """Get the viewers for this file format.

        This may expand to accommodate additional file formats or alternate
        methods.
        """
        try:
            viewers = self._viewer_map[path.suffix]
        except KeyError:
            TypeError(f"Unrecognized file type for {path}")
        else:
            return viewers

    def __getitem__(self, group: str):
        """Get the appropriate viewer for this dataset group."""
        if group in self._viewers:
            viewer = self._viewers[group]
            return viewer(self.path)
        raise KeyError(f"No viewer for group '{group}' in {self.path}")

    def __setitem__(self, group: str, viewer: typing.Type[Viewer]):
        """Associate a new viewer with this dataset group."""
        self._viewers[group] = viewer

    def __delitem__(self, group: str):
        """Delete the viewer associated with this dataset group."""
        del self._viewers[group]


class SubsetKeys(typing.NamedTuple):
    """A subset of names of attributes."""

    full: typing.Tuple[str]
    """Every attribute's names and aliases in a flat sequence."""
    aliased: typing.Tuple[aliased.MappingKey]
    """Groups of aliases for each attribute."""
    canonical: typing.Tuple[str]
    """The (unaliased) name of each attribute as defined on disk."""


class Interface(iterables.ReprStrMixin, metaclass=iotools.PathSet):
    """A format-agnostic view of a dataset.
    
    An instance of this class provides aliased access to variables and axes
    defined in a specific dataset, given a path to that dataset. It is designed
    to provide a single interface, regardless of file type, with as little
    overhead as possible. Therefore, it does not attempt to modify attributes
    (e.g., converting variable units), since doing so could result in reading a
    potentially large array from disk.
    """

    def __init__(self, path: iotools.PathLike) -> None:
        self.path = iotools.ReadOnlyPath(path)
        self.viewers = ViewerFactory(self.path)
        self._variables = None
        self._axes = None
        self._units = None

    @property
    def variables(self):
        """The variables in this dataset."""
        if self._variables is None:
            variables = {
                reference.ALIASES.get(name, name): variable
                for name, variable in self.viewers['variables'].items()
            }
            self._variables = aliased.Mapping(variables)
        return self._variables

    @property
    def units(self):
        """The unit of each variable, if available."""
        if self._units is None:
            units = {
                name: variable.unit
                for name, variable in self.variables.items(aliased=True)
            }
            self._units = aliased.Mapping(units)
        return self._units

    @property
    def axes(self):
        """The axes in this dataset."""
        if self._axes is None:
            axes = {
                reference.ALIASES.get(name, name): axis
                for name, axis in self.viewers['axes'].items()
            }
            self._axes = aliased.Mapping(axes)
        return self._axes

    def available(self, key: str):
        """Provide the names of available attributes."""
        if key in {'variable', 'variables'}:
            return SubsetKeys(
                full=tuple(self.variables),
                aliased=tuple(self.variables.keys(aliased=True)),
                canonical=tuple(self.viewers['variables'].keys()),
            )
        if key in {'axis', 'axes'}:
            return SubsetKeys(
                full=tuple(self.axes),
                aliased=tuple(self.axes.keys(aliased=True)),
                canonical=tuple(self.viewers['axes'].keys()),
            )

    def use(self, **viewers) -> 'Interface':
        """Update the viewers for this instance."""
        self.viewers.update(viewers)
        return self

    def __str__(self) -> str:
        return str(self.path)





import typing

from goats.core import axis
from goats.core import datafile
from goats.core import metric
from goats.core import variable


class Interface:
    """"""

    def __init__(
        self,
        __datafile: datafile.Interface,
        system: typing.Union[str, metric.System]=None,
    ) -> None:
        self._datafile = __datafile
        self._system = system
        self._axes = None
        self._variables = None

    @property
    def axes(self):
        """An interface to this dataset's axis quantities."""
        if self._axes is None:
            self._axes = axis.Interface(self._datafile, self.system)
        return self._axes

    @property
    def variables(self):
        """An interface to this dataset's variable quantities."""
        if self._variables is None:
            self._variables = variable.Interface(self._datafile, self.system)
        return self._variables

    @property
    def system(self):
        """This dataset's metric system."""
        return self._system



import typing

from goats.core import symbolic
from goats.core import iterables
from goats.core import metadata
from goats.core import metric
from goats.core import observing


def iscomposed(this):
    """True if `this` is an symbolic composition of observable quantities.
    
    Parameters
    ----------
    this
        The object to check.

    Notes
    -----
    This is more stringent than simply checking whether `this` can instantiate
    an `~symbolic.Expression` because all names of observable quantities would
    satisfy that condition.
    """
    return (
        isinstance(this, symbolic.Expression)
        or isinstance(this, str) and ('/' in this or '*' in this)
    )


class Quantity(iterables.ReprStrMixin):
    """A quantity that produces an observation."""

    def __init__(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
        implementation: observing.Implementation,
        application: observing.Application,
        unit: metadata.UnitLike=None,
    ) -> None:
        """
        Initialize this instance.

        Parameters
        ----------
        name : string, iterable of strings, or `~metadata.Name`
            The name(s) of this observable quantity.

        implementation : `~observing.Implementation`
            The object that represents the observing interface to the target
            observable quantity.

        application : `~observing.Application`
            An existing observing application that will compute the observed
            quantity, manage user constraints, and provide default parameter
            values.

        unit : unit-like, optional
            The metric unit to which to convert observations of this quantity.
        """
        self._name = metadata.Name(name)
        self._implementation = implementation
        self.application = application
        self._unit = metadata.Unit(unit or self._implementation.get_unit(name))
        self._axes = None

    def __getitem__(self, __x: metadata.UnitLike):
        """Create a quantity with the new unit."""
        unit = (
            self.unit.norm[__x]
            if str(__x).lower() in metric.SYSTEMS else __x
        )
        if unit == self._unit:
            return self
        return Quantity(
            self.name,
            self._implementation,
            self.application,
            unit=metadata.Unit(unit),
        )

    @property
    def unit(self):
        """This quantity's current metric unit."""
        return self._unit

    @property
    def axes(self):
        """This quantity's indexable axes."""
        if self._axes is None:
            self._axes = self._implementation.get_axes(self.name)
        return self._axes

    @property
    def name(self):
        """This quantity's name."""
        return self._name

    def observe(self, **constraints):
        """Observe this observable quantity."""
        self.application.apply(**constraints)
        result = self._implementation.apply(self.application, unit=self.unit)
        self.application.reset()
        return result

    def __eq__(self, __o) -> bool:
        """True if two instances have equivalent attributes."""
        if isinstance(__o, Quantity):
            return all(
                getattr(self, attr) == getattr(__o, attr)
                for attr in ('name', 'unit', 'axes')
            )
        return NotImplemented

    def __str__(self) -> str:
        display = [
            f"{str(self.name)!r}",
            f"unit={str(self.unit)!r}",
            f"axes={str(self.axes)}",
        ]
        return ', '.join(display)




import typing

from goats.core import metadata
from goats.core import metric
from goats.core import observable
from goats.core import observing
from goats.core import reference
from goats.core import spelling
from goats.core import symbolic


class Interface:
    """The base class for all observers."""

    _data: observing.Interface=None

    def __init__(
        self,
        *unobservable: str,
        system: str='mks',
        apply: typing.Type[observing.Application]=None,
    ) -> None:
        """Initialize this instance.
        
        Parameters
        ----------
        *unobservable : string
            The names of variable quantities from this observer's dataset to
            exclude from the set of formally observable quantities. These
            quantities will still be accessible in their variable form.

        system : string, default='mks'
            The metric system to use for variable and observable quantities.

        application : type of observing application, optional
            A subclass of `~observing.Application` with which to evaluate
            operational parameters and variable quantities during the observing
            process. If `application` is absent, this class will use an instance
            of the base class.
        """
        self._unobservable = unobservable
        self._system = metric.System(system)
        self._application_type = apply or observing.Application
        self._source = None
        self._application = None
        self._spellcheck = None

    # TODO: I'm no longer sure it makes sense to allow the user to update an
    # observer's metric system rather than create a new observer for a new
    # metric system.
    def system(self, new: str=None):
        """Get or set this observer's metric system."""
        if not new:
            return self._system
        self._system = metric.System(new)
        return self

    def readfrom(self, source):
        """Update this observer's data source."""
        self._source = source
        self._application = None
        return self

    @property
    def source(self):
        """The source of this observer's data."""
        return self._source

    @property
    def application(self):
        """This observer's observing application."""
        if self._application is None:
            self._application = self._application_type(self.data)
        return self._application

    @property
    def assumptions(self):
        """The operational assumptions available to this observer."""
        return self.data.assumptions

    @property
    def observables(self):
        """The names of formally observable quantities.
        
        This property contains the names of all primary and derived observable
        quantities. A primary derived observable quantity is one that comes
        directly from this observer's dataset; a derived observable quantity is
        one that is the result of a defined function.

        Note that it is also possible to symbolically compose new observable
        quantities from those listed here. Therefore, this collection represents
        the minimal set of quantities that this observer can observe.
        """
        fromdata = list(self.data.variables) + list(self.data.functions)
        available = set(fromdata) - set(self._unobservable)
        return tuple(available)

    @property
    def data(self):
        """An interface to this observer's dataset quantities."""
        if self._data is None:
            raise NotImplementedError(
                f"Observer requires an observing interface"
            ) from None
        return self._data

    def __getitem__(self, key: str):
        """Access an observable quantity by keyword, if possible."""
        if q := self._get_quantity(key):
            return q
        self._check_spelling(key) # -> None if `key` is spelled correctly
        raise KeyError(f"No observable for {key!r}") from None

    def _get_quantity(self, key: str):
        """Retrieve the named quantity from an interface, if possible."""
        if self.knows(key):
            name = metadata.Name(
                key if observable.iscomposed(key)
                else reference.ALIASES[key]
            )
            return observable.Quantity(
                name,
                observing.Implementation(key, self.data),
                application=self.application,
            )
        if key in self.data.variables:
            return self.data.variables[key]
        if key in self.assumptions:
            return self.assumptions[key]

    def knows(
        self,
        name: typing.Union[str, typing.Iterable[str], metadata.Name],
    ) -> bool:
        """True if this interface can observe the named quantity."""
        if isinstance(name, str):
            return self._knows(name)
        return next((self._knows(key) for key in name), False)

    def _knows(self, key: str):
        """Internal helper for `~Interface.knows`."""
        if key in self.observables:
            return True
        if observable.iscomposed(key):
            expression = symbolic.Expression(key)
            return all(term.base in self.observables for term in expression)
        return False

    def _check_spelling(self, key: str):
        """Catch misspelled names of observable quantities, if possible."""
        keys = list(self.observables) + list(self.assumptions)
        if self._spellcheck is None:
            self._spellcheck = spelling.SpellChecker(*keys)
        else:
            self._spellcheck.words |= keys
        return self._spellcheck.check(key)




import typing


class SpellingError(Exception):
    """It seems like the user just misspelled the variable's name."""
    def __init__(
        self,
        name: str,
        suggested: typing.Union[str, typing.List[str]],
    ) -> None:
        self.name = name
        self.suggested = suggested

    def __str__(self) -> str:
        return f"Could not find '{self.name}'. {self.suggestion}"

    @property
    def suggestion(self) -> str:
        """The formatted spelling suggestion."""
        if isinstance(self.suggested, str):
            return f"Did you mean '{self.suggested}'?"
        if len(self.suggested) == 1:
            return f"Did you mean '{self.suggested[0]}'?"
        return f"Did you mean one of {self.suggested}?"


class SpellChecker:
    """A simple spell-checker for named quantities.
    
    This is based on https://norvig.com/spell-correct.html
    """
    def __init__(self, *words: str) -> None:
        self.words = set(words)
        letters = 'abcdefghijklmnopqrstuvwxyz'
        self.letters = letters + letters.upper()

    def check(self, name: str, mode: str=None):
        """Check the spelling of `name` based on known words."""
        suggestions = self._check(name)
        if mode == 'suggest':
            return suggestions
        if mode == 'truth':
            return not suggestions
        if mode is not None:
            raise ValueError(f"Unknown mode {mode!r}") from None
        if suggestions:
            raise SpellingError(name, suggestions)

    def _check(self, name: str):
        """Internal helper for `~SpellChecker.check`."""
        if name in self.words:
            return []
        edits = list(self.edits(name))
        suggestions = self.get_suggestions(edits)
        if not suggestions:
            edits = [w for edit in edits for w in list(self.edits(edit))]
            suggestions = self.get_suggestions(edits)
        return suggestions

    def get_suggestions(self, edits: list):
        """Get a batch of suggested words based on `edits`."""
        return [n for n in self.words if n in edits]

    def known(self, words: typing.Iterable[str]) -> set:
        """The subset of `words` that is in the list of known words."""
        return {word for word in words if word in self.words}

    def edits(self, word: str) -> typing.Set[typing.List[str]]:
        """All edits that are one edit away from `word`."""
        return set(
            self.deletes(word)
            + self.transposes(word)
            + self.replaces(word)
            + self.inserts(word)
        )

    def splits(self, word: str) -> typing.List[str]:
        """The strings made by splitting `word` at each pair of letters."""
        return [
            (word[:i], word[i:])
            for i in range(len(word))
        ]

    def deletes(self, word: str) -> typing.List[str]:
        """All words produced by deleting one letter from `word`."""
        return [
            left + right[1:]
            for left, right in self.splits(word) if right
        ]

    def transposes(self, word: str) -> typing.List[str]:
        """All words produced by transposing one pair of letters in `word`."""
        return [
            left + right[1] + right[0] + right[2:]
            for left, right in self.splits(word) if len(right) > 1
        ]

    def replaces(self, word: str) -> typing.List[str]:
        """All words produced by replacing one letter in `word`."""
        return [
            left + c + right[1:] 
            for left, right in self.splits(word) if right for c in self.letters
        ]

    def inserts(self, word: str) -> typing.List[str]:
        """All words produced by inserting one letter into `word`."""
        return [
            left + c + right
            for left, right in self.splits(word) for c in self.letters
        ]





from goats.core import aliased


__all__ = [
    'ALIASES',
    'METADATA',
    'NAMES',
]


_BASE = {
    ('time', 't', 'times'): {
        'quantity': 'time',
    },
    ('shell', 'shells'): {
        'quantity': 'number',
    },
    (
        'mu', 'mus',
        'pitch angle', 'pitch-angle', 'pitch-angle cosine',
        'pitch angles', 'pitch-angles', 'pitch-angle cosines',
    ): {
        'quantity': 'ratio',
    },
    ('mass', 'm'): {
        'quantity': 'mass',
    },
    ('charge', 'q'): {
        'quantity': 'charge',
    },
    ('energy', 'egrid', 'energies', 'E'): {
        'quantity': 'energy',
    },
    ('speed', 'vgrid', 'vparticle'): {
        'quantity': 'velocity',
    },
    ('radius', 'R', 'r'): {
        'quantity': 'length',
    },
    ('theta', 'T'): {
        'quantity': 'plane angle',
    },
    ('phi', 'P'): {
        'quantity': 'plane angle',
    },
    ('Br', 'br'): {
        'quantity': 'magnetic field',
    },
    ('Bt', 'bt', 'Btheta', 'btheta'): {
        'quantity': 'magnetic field',
    },
    ('Bp', 'bp', 'Bphi', 'bphi'): {
        'quantity': 'magnetic field',
    },
    ('Vr', 'vr'): {
        'quantity': 'velocity',
    },
    ('Vt', 'vt', 'Vtheta', 'vtheta'): {
        'quantity': 'velocity',
    },
    ('Vp', 'vp', 'Vphi', 'vphi'): {
        'quantity': 'velocity',
    },
    ('rho', 'Rho'): {
        'quantity': 'number density',
    },
    ('dist', 'Dist', 'f'): {
        'quantity': 'particle distribution',
    },
    ('flux', 'Flux', 'J', 'J(E)', 'j', 'j(E)'): {
        'quantity': (
            'number / (area * solid_angle * time * energy / mass_number)'
        ),
    },
    ('x', 'X'): {
        'quantity': 'length',
    },
    ('y', 'Y'): {
        'quantity': 'length',
    },
    ('z', 'Z'): {
        'quantity': 'length',
    },
    ('B', 'b_mag', '|B|', 'bmag', 'b mag'): {
        'quantity': 'magnetic field',
    },
    ('V', 'v_mag', '|V|', 'vmag', 'v mag', 'v', '|v|'): {
        'quantity': 'velocity',
    },
    ('BV', 'bv_mag', 'bv', '|bv|', '|BV|'): {
        'quantity': 'velocity * magnetic field',
    },
    ('Vpara', 'v_para', 'vpara'): {
        'quantity': 'velocity',
    },
    ('Vperp', 'v_perp', 'vperp'): {
        'quantity': 'velocity',
    },
    ('flow_angle', 'flow angle', 'angle'): {
        'quantity': 'plane angle',
    },
    ('div_v', 'divV', 'divv', 'div V', 'div v', 'div(V)', 'div(v)'): {
        'quantity': '1 / time',
    },
    ('density_ratio', 'density ratio' ,'n2/n1', 'n_2/n_1'): {
        'quantity': 'number',
    },
    ('rigidity', 'Rg', 'R_g'): {
        'quantity': 'momentum / charge',
    },
    ('mean_free_path', 'mean free path', 'mfp'): {
        'quantity': 'length',
    },
    ('acceleration_rate', 'acceleration rate'): {
        'quantity': '1 / time',
    },
    ('energy_density', 'energy density'): {
        'quantity': 'energy / volume',
    },
    ('average_energy', 'average energy'): {
        'quantity': 'energy',
    },
    ('isotropic_distribution', 'isotropic distribution', 'isodist'): {
        'removed axes': ['mu'],
        'quantity': 'particle distribution',
    },
    'fluence': {
        'removed axes': ['time'],
        'quantity': 'number / (area * solid_angle * energy / mass_number)',
    },
    ('integral_flux', 'integral flux'): {
        'removed axes': ['energy'],
        'quantity': 'number / (area * solid_angle * time)',
    },
}

_NAMES = {
    v.get('name', k if isinstance(k, str) else k[0]): k
    for k, v in _BASE.items()
}

NAMES = aliased.NameMap(_NAMES)

_METADATA = {
    k: {i: j for i, j in v.items() if i != 'name'}
    for k, v in _BASE.items()
}

ALIASES = aliased.KeyMap(*_METADATA.keys())
METADATA = aliased.Mapping(_METADATA)

